from enum import Enum


class DatasetFileType(str, Enum):
    JSON = "json"
    CSV = "csv"
    PARQUET = "parquet"


class DatasetJoinKind(str, Enum):
    INNER = "inner"
    LEFT_OUTER = "left_outer"
    OUTER = "outer"
    RIGHT_OUTER = "right_outer"



import inspect
import logging
import typing
import uuid
from typing import Any, Callable, get_type_hints

from arthur_common.aggregations import (
    AggregationFunction,
    NumericAggregationFunction,
    SketchAggregationFunction,
)
from arthur_common.models.metrics import (
    AggregationMetricType,
    AggregationSpecSchema,
    DatasetReference,
    MetricsColumnListParameterSchema,
    MetricsColumnParameterSchema,
    MetricsDatasetParameterSchema,
    MetricsLiteralParameterSchema,
    MetricsParameterSchemaUnion,
)
from arthur_common.models.schema_definitions import (
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricLiteralParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    MetricsParameterAnnotationUnion,
)

logger = logging.getLogger(__name__)


class FunctionAnalyzer:
    @staticmethod
    def _python_type_to_scope_dtype(t: Any) -> DType:
        if t is int:
            return DType.INT
        elif t is str:
            return DType.STRING
        elif t is bool:
            return DType.BOOL
        elif t is float:
            return DType.FLOAT
        elif t is uuid.UUID:
            return DType.UUID
        elif t is DatasetReference:
            return DType.UUID
        elif typing.get_origin(t) is list:
            return DType.JSON
        elif typing.get_origin(t) is typing.Union:
            # handle union types to add support for Optional types only
            # extract the non-None type from Optional[T] = Union[T, None]
            args = typing.get_args(t)
            non_none_args = [arg for arg in args if arg is not type(None)]
            if len(non_none_args) == 1:
                return FunctionAnalyzer._python_type_to_scope_dtype(non_none_args[0])
            else:
                raise ValueError(
                    f"Union type {t} is not supported (only Optional[T] is supported).",
                )
        else:
            raise ValueError(f"Parameter type {t} is not supported.")

    @staticmethod
    def _get_metric_annotation_from_annotated(
        param_name: str,
        annotation: typing.Annotated,  # type: ignore
    ) -> MetricsParameterAnnotationUnion | None:
        arthur_metric_annotations = [
            m
            for m in annotation.__metadata__
            if isinstance(m, MetricsParameterAnnotationUnion)
        ]
        metric_annotation: MetricsParameterAnnotationUnion | None = None
        if len(arthur_metric_annotations) == 1:
            metric_annotation = arthur_metric_annotations[0]
        if len(arthur_metric_annotations) > 1:
            raise ValueError(
                f"Parameter {param_name} defines more than one metric annotation.",
            )
        return metric_annotation

    @staticmethod
    def _get_scope_metric_parameter_from_annotation(
        param_name: str,
        param_dtype: DType,
        optional: bool,
        annotation: typing.Annotated,  # type: ignore
    ) -> MetricsParameterSchemaUnion:
        if annotation is None:
            return MetricsLiteralParameterSchema(
                parameter_key=param_name,
                optional=optional,
                parameter_dtype=param_dtype,
                friendly_name=param_name,
                description=f"A {param_dtype.value} value.",
            )
        elif isinstance(annotation, MetricLiteralParameterAnnotation):
            return MetricsLiteralParameterSchema(
                parameter_key=param_name,
                optional=optional,
                parameter_dtype=param_dtype,
                friendly_name=annotation.friendly_name,
                description=annotation.description,
            )
        elif isinstance(annotation, MetricDatasetParameterAnnotation):
            if param_dtype != DType.UUID:
                raise ValueError(
                    f"Dataset parameter {param_name} has type {param_dtype}, but should be a UUID.",
                )
            return MetricsDatasetParameterSchema(
                parameter_key=param_name,
                optional=optional,
                friendly_name=annotation.friendly_name,
                description=annotation.description,
                model_problem_type=annotation.model_problem_type,
            )
        elif isinstance(annotation, MetricMultipleColumnParameterAnnotation):
            if param_dtype != DType.JSON:
                raise ValueError(
                    f"Dataset parameter {param_name} has type {param_dtype}, but should be a JSON type (valid list expected).",
                )
            return MetricsColumnListParameterSchema(
                parameter_key=param_name,
                tag_hints=annotation.tag_hints,
                optional=optional,
                source_dataset_parameter_key=annotation.source_dataset_parameter_key,
                allowed_column_types=annotation.allowed_column_types,
                allow_any_column_type=annotation.allow_any_column_type,
                friendly_name=annotation.friendly_name,
                description=annotation.description,
            )
        elif isinstance(annotation, MetricColumnParameterAnnotation):
            if param_dtype != DType.STRING:
                raise ValueError(
                    f"Column parameter {param_name} has type {param_dtype}, but should be a string.",
                )
            return MetricsColumnParameterSchema(
                parameter_key=param_name,
                tag_hints=annotation.tag_hints,
                optional=optional,
                source_dataset_parameter_key=annotation.source_dataset_parameter_key,
                allowed_column_types=annotation.allowed_column_types,
                allow_any_column_type=annotation.allow_any_column_type,
                friendly_name=annotation.friendly_name,
                description=annotation.description,
            )
        else:
            raise ValueError(
                f"Parameter {param_name} has an unsupported annotation {annotation}.",
            )

    """
    Returns a list of parameter names, parameter types, scope-specific annotations.
    """

    @staticmethod
    def _extract_parameter_metadata(func: Callable) -> list[MetricsParameterSchemaUnion]:  # type: ignore
        parameter_schemas: list[MetricsParameterSchemaUnion] = []
        args = inspect.signature(func).parameters
        for name, param in args.items():
            if name == "self":
                continue
            if name == "ddb_conn":
                continue

            if param.annotation == inspect.Parameter.empty:
                raise ValueError(
                    f"{func.__name__} must provide type annotation for parameter {name}.",
                )
            parameter_schemas.append(
                FunctionAnalyzer._get_scope_metric_parameter_from_annotation(
                    name,
                    FunctionAnalyzer._python_type_to_scope_dtype(
                        get_type_hints(func)[name],
                    ),
                    param.default != inspect.Parameter.empty,
                    (
                        FunctionAnalyzer._get_metric_annotation_from_annotated(
                            name,
                            param.annotation,
                        )
                        if typing.get_origin(param.annotation) is typing.Annotated
                        else None
                    ),
                ),
            )

        return parameter_schemas

    @staticmethod
    def analyze_aggregation_function(agg_func: type) -> AggregationSpecSchema:
        # Check if X is a subclass of AggregationFunction
        if not issubclass(agg_func, AggregationFunction):
            raise TypeError(
                f"Class {agg_func.__name__} is not a subclass of AggregationFunction.",
            )

        if issubclass(agg_func, NumericAggregationFunction):
            metric_type = AggregationMetricType.NUMERIC
        elif issubclass(agg_func, SketchAggregationFunction):
            metric_type = AggregationMetricType.SKETCH
        else:
            raise ValueError(
                f"Class {agg_func.__name__} is not a subclass of SketchAggregationFunction, NumericAggregationFunction.",
            )
        # Check if X implements the required methods
        required_methods = ["aggregate", "id", "description", "display_name"]
        static_methods = ["description", "id", "display_name", "reported_aggregations"]
        for method in required_methods:
            if not hasattr(agg_func, method) or not callable(getattr(agg_func, method)):
                raise AttributeError(
                    f"Class {agg_func.__name__} does not implement {method} method.",
                )

        for method in static_methods:
            if not is_static_method(getattr(agg_func, method)):
                raise AttributeError(f"Method {method} should be a staticmethod.")
        # Check if X passes the ABC implementation:
        try:
            agg_func()
        except TypeError as e:
            if "Can't instantiate abstract class" in str(e):
                logger.error(str(e))
                raise TypeError(
                    f"Class {agg_func.__name__} does not implement all the base class functions.",
                )
            else:
                # This is okay, it just means we didn't supply proper args to the __init__ function. The ABC mismatch would throw before this, so it must have passed
                pass

        aggregation_init_args: list[MetricsParameterSchemaUnion] = []

        # This is necessary because all the way down in the ABC class, some __init__ function is defined which we don't care about. Users should be able to exclude an init function and this allows them to do that.
        if has_custom_init(agg_func):
            aggregation_init_args = FunctionAnalyzer._extract_parameter_metadata(
                agg_func.__init__,
            )
        aggregate_args = FunctionAnalyzer._extract_parameter_metadata(
            agg_func.aggregate,
        )

        aggregation_id = agg_func.id()
        aggregation_description = agg_func.description()

        return AggregationSpecSchema(
            name=agg_func.display_name(),
            id=aggregation_id,
            # TODO: Require description, version
            description=aggregation_description,
            # version=0,
            metric_type=metric_type,
            init_args=aggregation_init_args,
            aggregate_args=aggregate_args,
            reported_aggregations=agg_func.reported_aggregations(),
        )


def has_custom_init(cls: type) -> bool:
    init_method = getattr(cls, "__init__", None)
    base_init_method = (
        getattr(cls.__base__, "__init__", None) if hasattr(cls, "__base__") else None
    )
    return init_method is not base_init_method


def is_static_method(method: type) -> bool:
    if inspect.isfunction(method):
        # Check if the method accepts no arguments or only default arguments
        argspec = inspect.getfullargspec(method)
        if len(argspec.args) == 0 and not argspec.varargs and not argspec.varkw:
            return True
    return False



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricLiteralParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)


class BinaryClassifierCountByClassAggregationFunction(NumericAggregationFunction):
    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000001f")

    @staticmethod
    def display_name() -> str:
        return "Binary Classification Count by Class - Class Label"

    @staticmethod
    def description() -> str:
        return "Aggregation that counts the number of predictions by class for a binary classifier. Takes boolean, integer, or string prediction values and groups them by time bucket to show prediction distribution over time."

    @staticmethod
    def _metric_name() -> str:
        return "binary_classifier_count_by_class"

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=BinaryClassifierCountByClassAggregationFunction._metric_name(),
                description=BinaryClassifierCountByClassAggregationFunction.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing binary classifier prediction values.",
                model_problem_type=ModelProblemType.BINARY_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.BOOL),
                    ScalarType(dtype=DType.INT),
                    ScalarType(dtype=DType.STRING),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing boolean, integer, or string labelled prediction values.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        """Executed SQL with no segmentation columns:
        SELECT
            time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts,
            {prediction_col} as prediction,
            COUNT(*) as count
        FROM {dataset.dataset_table_name}
        GROUP BY
            ts,
            -- group by raw column name instead of alias in select
            -- in case table has a column called 'prediction'
            {prediction_col}
        ORDER BY ts
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts",
            f"{prediction_col} as prediction",
            f"COUNT(*) as count",
        ] + segmentation_cols
        all_group_by_cols = ["ts", f"{prediction_col}"] + segmentation_cols
        extra_dims = ["prediction"]

        # build query
        query = f"""
                    SELECT {", ".join(all_select_clause_cols)}
                    FROM {dataset.dataset_table_name}
                    GROUP BY {", ".join(all_group_by_cols)}
                    ORDER BY ts
                """

        result = ddb_conn.sql(query).df()

        series = self.group_query_results_to_numeric_metrics(
            result,
            "count",
            segmentation_cols + extra_dims,
            "ts",
        )
        metric = self.series_to_metric(self._metric_name(), series)
        return [metric]


class BinaryClassifierCountThresholdClassAggregationFunction(
    NumericAggregationFunction,
):
    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000020")

    @staticmethod
    def display_name() -> str:
        return "Binary Classification Count by Class - Probability Threshold"

    @staticmethod
    def description() -> str:
        return "Aggregation that counts the number of predictions by class for a binary classifier using a probability threshold. Takes float prediction values and a threshold value to classify predictions, then groups them by time bucket to show prediction distribution over time."

    @staticmethod
    def _metric_name() -> str:
        return "binary_classifier_count_by_class"

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=BinaryClassifierCountThresholdClassAggregationFunction._metric_name(),
                description=BinaryClassifierCountThresholdClassAggregationFunction.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing binary classifier prediction values.",
                model_problem_type=ModelProblemType.BINARY_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing float prediction values.",
            ),
        ],
        threshold: Annotated[
            float,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.FLOAT,
                friendly_name="Threshold",
                description="The threshold to classify predictions to 0 or 1. 0 will result in the 'False Label' being assigned and 1 to the 'True Label' being assigned.",
            ),
        ],
        true_label: Annotated[
            str,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.STRING,
                friendly_name="True Label",
                description="The label denoting a positive classification.",
            ),
        ],
        false_label: Annotated[
            str,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.STRING,
                friendly_name="False Label",
                description="The label denoting a negative classification.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        """Executed SQL with no segmentation columns:
            SELECT
            time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts,
            CASE WHEN {prediction_col} >= {threshold} THEN '{true_label}' ELSE '{false_label}' END as prediction,
            COUNT(*) as count
        FROM {dataset.dataset_table_name}
        GROUP BY
            ts,
            -- group by raw column name instead of alias in select
            -- in case table has a column called 'prediction'
            {prediction_col}
        ORDER BY ts
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts",
            f"CASE WHEN {prediction_col} >= {threshold} THEN '{true_label}' ELSE '{false_label}' END as prediction",
            f"COUNT(*) as count",
        ] + segmentation_cols
        all_group_by_cols = [
            "ts",
            f"{prediction_col}",
        ] + segmentation_cols
        extra_dims = ["prediction"]

        query = f"""
            SELECT {", ".join(all_select_clause_cols)}
            FROM {dataset.dataset_table_name}
            GROUP BY {", ".join(all_group_by_cols)}
            ORDER BY ts
        """

        result = ddb_conn.sql(query).df()

        series = self.group_query_results_to_numeric_metrics(
            result,
            "count",
            segmentation_cols + extra_dims,
            "ts",
        )
        metric = self.series_to_metric(self._metric_name(), series)
        return [metric]



from datetime import datetime
from typing import Any, Dict, List, Optional, Self, Type

from fastapi import HTTPException
from openinference.semconv.trace import OpenInferenceSpanKindValues
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)

from arthur_common.models.common_schemas import (
    ExamplesConfig,
    KeywordsConfig,
    PIIConfig,
    RegexConfig,
    ToxicityConfig,
)
from arthur_common.models.constants import (
    ERROR_PASSWORD_POLICY_NOT_MET,
    GENAI_ENGINE_KEYCLOAK_PASSWORD_LENGTH,
    HALLUCINATION_RULE_NAME,
    NEGATIVE_BLOOD_EXAMPLE,
)
from arthur_common.models.enums import (
    APIKeysRolesEnum,
    InferenceFeedbackTarget,
    MetricType,
    PIIEntityTypes,
    RuleScope,
    RuleType,
    StatusCodeEnum,
    ToolClassEnum,
    ContinuousEvalRunStatus,
    AgenticAnnotationType,
)
from arthur_common.models.metric_schemas import RelevanceMetricConfig


class UpdateRuleRequest(BaseModel):
    enabled: bool = Field(description="Boolean value to enable or disable the rule. ")


# Using the latest version from arthur-common
class NewRuleRequest(BaseModel):
    name: str = Field(description="Name of the rule", examples=["SSN Regex Rule"])
    type: str = Field(
        description="Type of the rule. It can only be one of KeywordRule, RegexRule, "
        "ModelSensitiveDataRule, ModelHallucinationRule, ModelHallucinationRuleV2, PromptInjectionRule, PIIDataRule",
        examples=["RegexRule"],
    )
    apply_to_prompt: bool = Field(
        description="Boolean value to enable or disable the rule for llm prompt",
        examples=[True],
    )
    apply_to_response: bool = Field(
        description="Boolean value to enable or disable the rule for llm response",
        examples=[False],
    )
    config: (
        KeywordsConfig
        | RegexConfig
        | ExamplesConfig
        | ToxicityConfig
        | PIIConfig
        | None
    ) = Field(description="Config of the rule", default=None)

    model_config = ConfigDict(
        json_schema_extra={
            "example1": {
                "summary": "Sensitive Data Example",
                "description": "Sensitive Data Example with its required configuration",
                "value": {
                    "name": "Sensitive Data Rule",
                    "type": "ModelSensitiveDataRule",
                    "apply_to_prompt": True,
                    "apply_to_response": False,
                    "config": {
                        "examples": [
                            {
                                "example": NEGATIVE_BLOOD_EXAMPLE,
                                "result": True,
                            },
                            {
                                "example": "Most of the people have A positive blood group",
                                "result": False,
                            },
                        ],
                        "hint": "specific individual's blood types",
                    },
                },
            },
            "example2": {
                "summary": "Regex Example",
                "description": "Regex Example with its required configuration. Be sure to properly encode requests "
                "using JSON libraries. For example, the regex provided encodes to a different string "
                "when encoded to account for escape characters.",
                "value": {
                    "name": "SSN Regex Rule",
                    "type": "RegexRule",
                    "apply_to_prompt": True,
                    "apply_to_response": True,
                    "config": {
                        "regex_patterns": [
                            "\\d{3}-\\d{2}-\\d{4}",
                            "\\d{5}-\\d{6}-\\d{7}",
                        ],
                    },
                },
            },
            "example3": {
                "summary": "Keywords Rule Example",
                "description": "Keywords Rule Example with its required configuration",
                "value": {
                    "name": "Blocked Keywords Rule",
                    "type": "KeywordRule",
                    "apply_to_prompt": True,
                    "apply_to_response": True,
                    "config": {"keywords": ["Blocked_Keyword_1", "Blocked_Keyword_2"]},
                },
            },
            "example4": {
                "summary": "Prompt Injection Rule Example",
                "description": "Prompt Injection Rule Example, no configuration required",
                "value": {
                    "name": "Prompt Injection Rule",
                    "type": "PromptInjectionRule",
                    "apply_to_prompt": True,
                    "apply_to_response": False,
                },
            },
            "example5": {
                "summary": "Hallucination Rule V1 Example (Deprecated)",
                "description": "Hallucination Rule Example, no configuration required (This rule is deprecated. Use "
                "ModelHallucinationRuleV2 instead.)",
                "value": {
                    "name": HALLUCINATION_RULE_NAME,
                    "type": "ModelHallucinationRule",
                    "apply_to_prompt": False,
                    "apply_to_response": True,
                },
            },
            "example6": {
                "summary": "Hallucination Rule V2 Example",
                "description": "Hallucination Rule Example, no configuration required",
                "value": {
                    "name": HALLUCINATION_RULE_NAME,
                    "type": "ModelHallucinationRuleV2",
                    "apply_to_prompt": False,
                    "apply_to_response": True,
                },
            },
            "example7": {
                "summary": "Hallucination Rule V3 Example (Beta)",
                "description": "Hallucination Rule Example, no configuration required. This rule is in beta and must "
                "be enabled by the system administrator.",
                "value": {
                    "name": HALLUCINATION_RULE_NAME,
                    "type": "ModelHallucinationRuleV3",
                    "apply_to_prompt": False,
                    "apply_to_response": True,
                },
            },
            "example8": {
                "summary": "PII Rule Example",
                "description": f'PII Rule Example, no configuration required. "disabled_pii_entities", '
                f'"confidence_threshold", and "allow_list" accepted. Valid value for '
                f'"confidence_threshold" is 0.0-1.0. Valid values for "disabled_pii_entities" '
                f"are {PIIEntityTypes.to_string()}",
                "value": {
                    "name": "PII Rule",
                    "type": "PIIDataRule",
                    "apply_to_prompt": True,
                    "apply_to_response": True,
                    "config": {
                        "disabled_pii_entities": [
                            "EMAIL_ADDRESS",
                            "PHONE_NUMBER",
                        ],
                        "confidence_threshold": "0.5",
                        "allow_list": ["arthur.ai", "Arthur"],
                    },
                },
            },
            "example9": {
                "summary": "Toxicity Rule Example",
                "description": "Toxicity Rule Example, no configuration required. Threshold accepted",
                "value": {
                    "name": "Toxicity Rule",
                    "type": "ToxicityRule",
                    "apply_to_prompt": True,
                    "apply_to_response": True,
                    "config": {"threshold": 0.5},
                },
            },
        },
    )

    @model_validator(mode="before")
    def set_config_type(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        config_type_to_class: Dict[str, Type[BaseModel]] = {
            RuleType.REGEX: RegexConfig,
            RuleType.KEYWORD: KeywordsConfig,
            RuleType.TOXICITY: ToxicityConfig,
            RuleType.PII_DATA: PIIConfig,
            RuleType.MODEL_SENSITIVE_DATA: ExamplesConfig,
        }

        config_type = values["type"]
        config_class = config_type_to_class.get(config_type)

        if config_class is not None:
            config_values = values.get("config")
            if config_values is None:
                if config_type in [RuleType.REGEX, RuleType.KEYWORD]:
                    raise HTTPException(
                        status_code=400,
                        detail="This rule must be created with a config parameter",
                    )
                config_values = {}
            if isinstance(config_values, BaseModel):
                config_values = config_values.model_dump()
            values["config"] = config_class(**config_values)
        return values

    @model_validator(mode="after")
    def check_prompt_or_response(self) -> Self:
        if (self.type == RuleType.MODEL_SENSITIVE_DATA) and (
            self.apply_to_response is True
        ):
            raise HTTPException(
                status_code=400,
                detail="ModelSensitiveDataRule can only be enabled for prompt. Please set the 'apply_to_response' "
                "field to false.",
            )
        if (self.type == RuleType.PROMPT_INJECTION) and (
            self.apply_to_response is True
        ):
            raise HTTPException(
                status_code=400,
                detail="PromptInjectionRule can only be enabled for prompt. Please set the 'apply_to_response' field "
                "to false.",
            )
        if (self.type == RuleType.MODEL_HALLUCINATION_V2) and (
            self.apply_to_prompt is True
        ):
            raise HTTPException(
                status_code=400,
                detail="ModelHallucinationRuleV2 can only be enabled for response. Please set the 'apply_to_prompt' "
                "field to false.",
            )
        if (self.apply_to_prompt is False) and (self.apply_to_response is False):
            raise HTTPException(
                status_code=400,
                detail="Rule must be either applied to the prompt or to the response.",
            )

        return self

    @model_validator(mode="after")
    def check_examples_non_null(self) -> Self:
        if self.type == RuleType.MODEL_SENSITIVE_DATA:
            config = self.config
            if (
                config is not None
                and isinstance(config, ExamplesConfig)
                and (config.examples is None or len(config.examples) == 0)
            ):
                raise HTTPException(
                    status_code=400,
                    detail="Examples must be provided to onboard a ModelSensitiveDataRule",
                )
        return self


class SearchTasksRequest(BaseModel):
    task_ids: Optional[list[str]] = Field(
        description="List of tasks to query for.",
        default=None,
    )
    task_name: Optional[str] = Field(
        description="Task name substring search string.",
        default=None,
    )
    is_agentic: Optional[bool] = Field(
        description="Filter tasks by agentic status. If not provided, returns both agentic and non-agentic tasks.",
        default=None,
    )


class SearchRulesRequest(BaseModel):
    rule_ids: Optional[list[str]] = Field(
        description="List of rule IDs to search for.",
        default=None,
    )
    rule_scopes: Optional[list[RuleScope]] = Field(
        description="List of rule scopes to search for.",
        default=None,
    )
    prompt_enabled: Optional[bool] = Field(
        description="Include or exclude prompt-enabled rules.",
        default=None,
    )
    response_enabled: Optional[bool] = Field(
        description="Include or exclude response-enabled rules.",
        default=None,
    )
    rule_types: Optional[list[RuleType]] = Field(
        description="List of rule types to search for.",
        default=None,
    )


class NewTaskRequest(BaseModel):
    name: str = Field(description="Name of the task.", min_length=1)
    is_agentic: bool = Field(
        description="Whether the task is agentic or not.",
        default=False,
    )


class NewApiKeyRequest(BaseModel):
    description: Optional[str] = Field(
        description="Description of the API key. Optional.",
        default=None,
    )
    roles: Optional[list[APIKeysRolesEnum]] = Field(
        description=f"Role that will be assigned to API key. Allowed values: {[role for role in APIKeysRolesEnum]}",
        default=[APIKeysRolesEnum.VALIDATION_USER],
    )


class PromptValidationRequest(BaseModel):
    prompt: str = Field(description="Prompt to be validated by GenAI Engine")
    # context: Optional[str] = Field(
    #     description="Optional data provided as context for the prompt validation. "
    #     "Currently not used"
    # )
    conversation_id: Optional[str] = Field(
        description="The unique conversation ID this prompt belongs to. All prompts and responses from this \
        conversation can later be reconstructed with this ID.",
        default=None,
    )
    user_id: Optional[str] = Field(
        description="The user ID this prompt belongs to",
        default=None,
    )


class ResponseValidationRequest(BaseModel):
    response: str = Field(description="LLM Response to be validated by GenAI Engine")
    context: Optional[str] = Field(
        description="Optional data provided as context for the validation.",
        default=None,
    )
    model_name: Optional[str] = Field(
        description="The model name and version being used for this response (e.g., 'gpt-4', 'gpt-3.5-turbo', 'claude-3-opus', 'gemini-pro').",
        default=None,
    )
    # tokens: Optional[List[str]] = Field(description="optional, not used currently")
    # token_likelihoods: Optional[List[str]] = Field(
    #     description="optional, not used currently"
    # )

    @model_validator(mode="after")
    def check_prompt_or_response(self) -> "ResponseValidationRequest":
        if isinstance(self, PromptValidationRequest) and self.prompt is None:
            raise ValueError("prompt is required when validating a prompt")
        if isinstance(self, ResponseValidationRequest) and self.response is None:
            raise ValueError("response is required when validating a response")
        return self


class ChatRequest(BaseModel):
    user_prompt: str = Field(description="Prompt user wants to send to chat.")
    conversation_id: str = Field(description="Conversation ID")
    file_ids: List[str] = Field(
        description="list of file IDs to retrieve from during chat.",
    )


class FeedbackRequest(BaseModel):
    target: InferenceFeedbackTarget
    score: int
    reason: str | None
    user_id: str | None = None


class CreateUserRequest(BaseModel):
    email: str
    password: str
    temporary: bool = True
    roles: list[str]
    firstName: str
    lastName: str


class PasswordResetRequest(BaseModel):
    password: str

    @field_validator("password")
    @classmethod
    def password_meets_security(cls, value: str) -> str:
        special_characters = '!@#$%^&*()-+?_=,<>/"'
        if not len(value) >= GENAI_ENGINE_KEYCLOAK_PASSWORD_LENGTH:
            raise ValueError(ERROR_PASSWORD_POLICY_NOT_MET)
        if (
            not any(c.isupper() for c in value)
            or not any(c.islower() for c in value)
            or not any(c.isdigit() for c in value)
            or not any(c in special_characters for c in value)
        ):
            raise ValueError(ERROR_PASSWORD_POLICY_NOT_MET)
        return value


class ChatDefaultTaskRequest(BaseModel):
    task_id: str


# Using the latest version from arthur-common
class NewMetricRequest(BaseModel):
    type: MetricType = Field(
        description="Type of the metric. It can only be one of QueryRelevance, ResponseRelevance, ToolSelection",
        examples=["UserQueryRelevance"],
    )
    name: str = Field(
        description="Name of metric",
        examples=["My User Query Relevance"],
    )
    metric_metadata: str = Field(description="Additional metadata for the metric")
    config: Optional[RelevanceMetricConfig] = Field(
        description="Configuration for the metric. Currently only applies to UserQueryRelevance and ResponseRelevance metric types.",
        default=None,
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example1": {
                "type": "QueryRelevance",
                "name": "My User Query Relevance",
                "metric_metadata": "This is a test metric metadata",
            },
            "example2": {
                "type": "QueryRelevance",
                "name": "My User Query Relevance with Config",
                "metric_metadata": "This is a test metric metadata",
                "config": {"relevance_threshold": 0.8, "use_llm_judge": False},
            },
            "example3": {
                "type": "ResponseRelevance",
                "name": "My Response Relevance",
                "metric_metadata": "This is a test metric metadata",
                "config": {"use_llm_judge": True},
            },
        },
    )

    @model_validator(mode="before")
    def set_config_type(cls, values: dict[str, Any] | None) -> dict[str, Any] | None:
        if not isinstance(values, dict):
            return values

        try:
            metric_type = MetricType(values.get("type", "empty_value"))
        except ValueError:
            raise HTTPException(
                status_code=400,
                detail=f"Invalid metric type: {values.get('type', 'empty_value')}. Must be one of {[t.value for t in MetricType]}",
                headers={"full_stacktrace": "false"},
            )

        config_values = values.get("config")

        # Map metric types to their corresponding config classes
        metric_type_to_config = {
            MetricType.QUERY_RELEVANCE: RelevanceMetricConfig,
            MetricType.RESPONSE_RELEVANCE: RelevanceMetricConfig,
            # Add new metric types and their configs here as needed
        }

        config_class = metric_type_to_config.get(metric_type)

        if config_class is not None:
            if config_values is None:
                # Default config when none is provided
                config_values = {"use_llm_judge": True}
            elif isinstance(config_values, dict):
                relevance_threshold = config_values.get("relevance_threshold")
                use_llm_judge = config_values.get("use_llm_judge")

                # Handle mutually exclusive parameters
                if relevance_threshold is not None and use_llm_judge:
                    raise HTTPException(
                        status_code=400,
                        detail="relevance_threshold and use_llm_judge=true are mutually exclusive. Set use_llm_judge=false when using relevance_threshold.",
                        headers={"full_stacktrace": "false"},
                    )

                # If relevance_threshold is set but use_llm_judge isn't, set use_llm_judge to false
                if relevance_threshold is not None and use_llm_judge is None:
                    config_values["use_llm_judge"] = False

                # If neither is set, default to use_llm_judge=True
                if relevance_threshold is None and (
                    use_llm_judge is None or use_llm_judge == False
                ):
                    config_values["use_llm_judge"] = True

            if isinstance(config_values, BaseModel):
                config_values = config_values.model_dump()

            values["config"] = config_class(**config_values)
        elif config_values is not None:
            # Provide a nice error message listing supported metric types
            supported_types = [t.value for t in metric_type_to_config.keys()]
            raise HTTPException(
                status_code=400,
                detail=f"Config is only supported for {', '.join(supported_types)} metric types",
                headers={"full_stacktrace": "false"},
            )

        return values


class UpdateMetricRequest(BaseModel):
    enabled: bool = Field(description="Boolean value to enable or disable the metric. ")


class SpanQueryRequest(BaseModel):
    """Request schema for querying spans with validation."""

    task_ids: list[str] = Field(
        ...,
        description="Task IDs to filter on. At least one is required.",
        min_length=1,
    )
    span_types: Optional[list[str]] = Field(
        None,
        description=f"Span types to filter on. Optional. Valid values: {', '.join(sorted([kind.value for kind in OpenInferenceSpanKindValues]))}",
    )
    start_time: Optional[datetime] = Field(
        None,
        description="Inclusive start date in ISO8601 string format.",
    )
    end_time: Optional[datetime] = Field(
        None,
        description="Exclusive end date in ISO8601 string format.",
    )
    session_ids: Optional[list[str]] = Field(
        None,
        description="Session IDs to filter on. Optional.",
    )
    span_ids: Optional[list[str]] = Field(
        None,
        description="Span IDs to filter on. Optional.",
    )
    user_ids: Optional[list[str]] = Field(
        None,
        description="User IDs to filter on. Optional.",
    )
    span_name: Optional[str] = Field(
        None,
        description="Return only results with this span name.",
    )
    span_name_contains: Optional[str] = Field(
        None,
        description="Return only results where span name contains this substring.",
    )
    status_code: Optional[list[StatusCodeEnum]] = Field(
        None,
        description="Status codes to filter on. Optional. Valid values: Ok, Error, Unset",
    )

    @field_validator("span_types")
    @classmethod
    def validate_span_types(cls, value: list[str]) -> list[str]:
        """Validate that all span_types are valid OpenInference span kinds."""
        if not value:
            return value

        # Get all valid span kind values
        valid_span_kinds = [kind.value for kind in OpenInferenceSpanKindValues]
        invalid_types = [st for st in value if st not in valid_span_kinds]

        if invalid_types:
            raise ValueError(
                f"Invalid span_types received: {invalid_types}. "
                f"Valid values: {', '.join(sorted(valid_span_kinds))}",
            )
        return value


class TraceQueryRequest(BaseModel):
    """Request schema for querying traces with comprehensive filtering."""

    # Required
    task_ids: list[str] = Field(
        ...,
        description="Task IDs to filter on. At least one is required.",
        min_length=1,
    )

    # Common optional filters
    trace_ids: Optional[list[str]] = Field(
        None,
        description="Trace IDs to filter on. Optional.",
    )
    start_time: Optional[datetime] = Field(
        None,
        description="Inclusive start date in ISO8601 string format. Use local time (not UTC).",
    )
    end_time: Optional[datetime] = Field(
        None,
        description="Exclusive end date in ISO8601 string format. Use local time (not UTC).",
    )

    # New trace-level filters
    tool_name: Optional[str] = Field(
        None,
        description="Return only results with this tool name.",
    )
    span_types: Optional[list[str]] = Field(
        None,
        description="Span types to filter on. Optional.",
    )
    span_ids: Optional[list[str]] = Field(
        None,
        description="Span IDs to filter on. Optional.",
    )
    session_ids: Optional[list[str]] = Field(
        None,
        description="Session IDs to filter on. Optional.",
    )
    user_ids: Optional[list[str]] = Field(
        None,
        description="User IDs to filter on. Optional.",
    )
    span_name: Optional[str] = Field(
        None,
        description="Return only results with this span name.",
    )
    span_name_contains: Optional[str] = Field(
        None,
        description="Return only results where span name contains this substring.",
    )
    status_code: Optional[list[StatusCodeEnum]] = Field(
        None,
        description="Status codes to filter on. Optional. Valid values: Ok, Error, Unset",
    )
    annotation_score: Optional[int] = Field(
        None,
        ge=0,
        le=1,
        description="Filter by trace annotation score (0 or 1).",
    )
    annotation_type: Optional[AgenticAnnotationType] = Field(
        None,
        description="Filter by trace annotation type (i.e. 'human' or 'continuous_eval').",
    )
    continuous_eval_run_status: Optional[ContinuousEvalRunStatus] = Field(
        None,
        description="Filter by trace annotation run status (e.g. 'passed', 'failed', etc.).",
    )
    continuous_eval_name: Optional[str] = Field(
        None,
        description="Filter by continuous eval name.",
    )

    # Query relevance filters
    query_relevance_eq: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Equal to this value.",
    )
    query_relevance_gt: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Greater than this value.",
    )
    query_relevance_gte: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Greater than or equal to this value.",
    )
    query_relevance_lt: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Less than this value.",
    )
    query_relevance_lte: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Less than or equal to this value.",
    )

    # Response relevance filters
    response_relevance_eq: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Equal to this value.",
    )
    response_relevance_gt: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Greater than this value.",
    )
    response_relevance_gte: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Greater than or equal to this value.",
    )
    response_relevance_lt: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Less than this value.",
    )
    response_relevance_lte: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Less than or equal to this value.",
    )

    # Tool classification filters
    tool_selection: Optional[ToolClassEnum] = Field(
        None,
        description="Tool selection evaluation result.",
    )
    tool_usage: Optional[ToolClassEnum] = Field(
        None,
        description="Tool usage evaluation result.",
    )

    # Trace duration filters
    trace_duration_eq: Optional[float] = Field(
        None,
        ge=0,
        description="Duration exactly equal to this value (seconds).",
    )
    trace_duration_gt: Optional[float] = Field(
        None,
        ge=0,
        description="Duration greater than this value (seconds).",
    )
    trace_duration_gte: Optional[float] = Field(
        None,
        ge=0,
        description="Duration greater than or equal to this value (seconds).",
    )
    trace_duration_lt: Optional[float] = Field(
        None,
        ge=0,
        description="Duration less than this value (seconds).",
    )
    trace_duration_lte: Optional[float] = Field(
        None,
        ge=0,
        description="Duration less than or equal to this value (seconds).",
    )

    @field_validator(
        "query_relevance_eq",
        "query_relevance_gt",
        "query_relevance_gte",
        "query_relevance_lt",
        "query_relevance_lte",
        "response_relevance_eq",
        "response_relevance_gt",
        "response_relevance_gte",
        "response_relevance_lt",
        "response_relevance_lte",
        mode="before",
    )
    @classmethod
    def validate_relevance_scores(
        cls,
        value: Optional[float],
        info: ValidationInfo,
    ) -> Optional[float]:
        """Validate that relevance scores are between 0 and 1 (inclusive)."""
        if value is not None:
            if not (0.0 <= value <= 1.0):
                raise ValueError(
                    f"{info.field_name} value must be between 0 and 1 (inclusive)",
                )
        return value

    @field_validator(
        "trace_duration_eq",
        "trace_duration_gt",
        "trace_duration_gte",
        "trace_duration_lt",
        "trace_duration_lte",
        mode="before",
    )
    @classmethod
    def validate_trace_duration(
        cls,
        value: Optional[float],
        info: ValidationInfo,
    ) -> Optional[float]:
        """Validate that trace duration values are non-negative."""
        if value is not None:
            if value < 0:
                raise ValueError(
                    f"{info.field_name} value must be non-negative (greater than or equal to 0)",
                )
        return value

    @field_validator("tool_selection", "tool_usage", mode="before")
    @classmethod
    def validate_tool_classification(cls, value: Any) -> Optional[ToolClassEnum]:
        """Validate tool classification enum values."""
        if value is not None:
            # Handle both integer and enum inputs
            if isinstance(value, int):
                if value not in [0, 1, 2]:
                    raise ValueError(
                        "Tool classification must be 0 (INCORRECT), "
                        "1 (CORRECT), or 2 (NA)",
                    )
                return ToolClassEnum(value)
            elif isinstance(value, ToolClassEnum):
                return value
            else:
                raise ValueError(
                    "Tool classification must be an integer (0, 1, 2) or ToolClassEnum instance",
                )
        return value

    @field_validator("span_types")
    @classmethod
    def validate_span_types(cls, value: Optional[list[str]]) -> Optional[list[str]]:
        """Validate that all span_types are valid OpenInference span kinds."""
        if not value:
            return value

        # Get all valid span kind values
        valid_span_kinds = [kind.value for kind in OpenInferenceSpanKindValues]
        invalid_types = [st for st in value if st not in valid_span_kinds]

        if invalid_types:
            raise ValueError(
                f"Invalid span_types received: {invalid_types}. "
                f"Valid values: {', '.join(sorted(valid_span_kinds))}",
            )
        return value

    @model_validator(mode="after")
    def validate_filter_combinations(self) -> Self:
        """Validate that filter combinations are logically valid."""
        # Check mutually exclusive filters for each metric type
        for prefix in ["query_relevance", "response_relevance", "trace_duration"]:
            eq_field = f"{prefix}_eq"
            comparison_fields = [f"{prefix}_{op}" for op in ["gt", "gte", "lt", "lte"]]

            if getattr(self, eq_field) and any(
                getattr(self, field) for field in comparison_fields
            ):
                raise ValueError(
                    f"{eq_field} cannot be combined with other {prefix} comparison operators",
                )

            # Check for incompatible operator combinations
            if getattr(self, f"{prefix}_gt") and getattr(self, f"{prefix}_gte"):
                raise ValueError(f"Cannot combine {prefix}_gt with {prefix}_gte")
            if getattr(self, f"{prefix}_lt") and getattr(self, f"{prefix}_lte"):
                raise ValueError(f"Cannot combine {prefix}_lt with {prefix}_lte")

        return self



import json
import re
from typing import Any

import duckdb
import pandas as pd
from dateutil.parser import parse
from fsspec import filesystem
from pydantic import BaseModel

from arthur_common.models.datasets import DatasetJoinKind
from arthur_common.models.schema_definitions import (
    DatasetListType,
    DatasetObjectType,
    DatasetScalarType,
    DatasetSchema,
    DType,
)

MAX_JSON_OBJECT_SIZE = 1024 * 1024 * 1024  # 1GB


class ColumnFormat(BaseModel):
    source_name: str
    alias: str
    format: str


class DuckDBOperator:
    """
    Loads data into a DuckDB table.

    If no schema is supplied, the output table will contain columns with names equal to the source names in the data.
    If a schema is applied, the column names in the output table will be aliases equal to the column id from the schema.
    This allows for consistent column naming across different data sources.
    """

    @staticmethod
    def load_data_to_duckdb(
        data: list[dict[str, Any]] | pd.DataFrame,
        preprocess_schema: bool = False,
        table_name: str = "inferences",
        conn: duckdb.DuckDBPyConnection | None = None,
        schema: DatasetSchema | None = None,
    ) -> duckdb.DuckDBPyConnection:
        if not conn:
            conn = duckdb.connect()

        if type(data) == list:
            DuckDBOperator._load_unstructured_data(data, table_name, conn, schema)
        elif type(data) == pd.DataFrame:
            DuckDBOperator._load_structured_data(
                data,
                preprocess_schema,
                table_name,
                conn,
                schema,
            )
        else:
            raise ValueError(f"Unsupported data type: {type(data)}")
        return conn

    """
    Rename columns from ids to their friendly names based on schema.column_names
    """

    @staticmethod
    def apply_alias_mask(
        table_name: str,
        conn: duckdb.DuckDBPyConnection,
        schema: DatasetSchema,
    ) -> None:
        old_new_mask = {
            str(col_id): schema.column_names[col_id] for col_id in schema.column_names
        }
        DuckDBOperator._apply_alias_mask(table_name, conn, old_new_mask)

    @staticmethod
    def _apply_alias_mask(
        table_name: str,
        conn: duckdb.DuckDBPyConnection,
        old_new_mask: dict[str, str],
    ) -> None:
        for old, new in old_new_mask.items():
            # Don't quote the join column names, since they're already quoted as part of escape_identifier's output
            alter_query = f"ALTER TABLE {table_name} RENAME COLUMN {escape_identifier(old)} TO {escape_identifier(new)}"
            conn.sql(alter_query)

    @staticmethod
    def _load_unstructured_data(
        data: list[dict[str, Any]],
        table_name: str,
        conn: duckdb.DuckDBPyConnection,
        schema: DatasetSchema | None,
    ) -> None:
        with filesystem("memory").open(f"inferences.json", "w") as file:
            file.write(json.dumps(data))
        conn.register_filesystem(filesystem("memory"))

        if schema:
            column_formats = make_duckdb_dataset_schema(schema)

            key_value_pairs = [
                f"{escape_identifier(col.source_name)}: '{col.format}'"
                for col in column_formats
            ]
            stringified_schema = ", ".join([f"{kv}" for kv in key_value_pairs])
            stringified_schema = f"{{ {stringified_schema} }}"

            read_stmt = f"read_json('memory://inferences.json', format='array', columns={stringified_schema}, maximum_object_size={MAX_JSON_OBJECT_SIZE})"
        else:
            read_stmt = f"read_json_auto('memory://inferences.json', maximum_object_size={MAX_JSON_OBJECT_SIZE})"

        conn.sql(
            f"CREATE OR REPLACE TEMP TABLE {table_name} AS SELECT * FROM {read_stmt}",
        )

        if schema:
            old_new_mask = {}
            for col in column_formats:
                old_new_mask[col.source_name] = col.alias
            DuckDBOperator._apply_alias_mask(table_name, conn, old_new_mask)

    @staticmethod
    def _load_structured_data(
        data: pd.DataFrame,
        preprocess_schema: bool,
        table_name: str,
        conn: duckdb.DuckDBPyConnection,
        schema: DatasetSchema | None,
    ) -> None:
        if preprocess_schema:
            data = DuckDBOperator._preprocess_dataframe_schema_inference(data)

        if schema:
            column_formats = make_duckdb_dataset_schema(schema)

            key_value_pairs = [
                f"{escape_identifier(col.source_name)} {col.format}"
                for col in column_formats
            ]
            stringified_schema = ", ".join([f"{kv}" for kv in key_value_pairs])
            create_table_stmt = (
                f"CREATE OR REPLACE TEMP TABLE {table_name} ({stringified_schema});"
            )
            conn.sql(create_table_stmt)
            conn.sql(f"INSERT INTO {table_name} SELECT * FROM data")

            old_new_mask = {}
            for col in column_formats:
                old_new_mask[col.source_name] = col.alias
            DuckDBOperator._apply_alias_mask(table_name, conn, old_new_mask)

        else:
            conn.sql(f"CREATE OR REPLACE TEMP TABLE {table_name} AS SELECT * FROM data")

    """
        Preprocess to make smarter type inferences. Pandas and json recognize very little beyond primitives out of the box. We can support a little more with a little effort like:
        1. Datetimes

        Modifies the input data in place to have smarter types than pandas will natively infer
    """

    @staticmethod
    def _preprocess_dataframe_schema_inference(data: pd.DataFrame) -> pd.DataFrame:
        datetime_columns = _infer_dataframe_datetime_columns(data)
        for column in datetime_columns:
            try:
                data[column] = pd.to_datetime(data[column])
            except Exception:
                # we're using best-effort to infer datetime columns, but just in case we got it wrong, move on
                continue

        return data

    @staticmethod
    def join_tables(
        conn: duckdb.DuckDBPyConnection,
        table_name: str,
        table_1: str,
        table_2: str,
        table_1_join_key: str,
        table_2_join_key: str,
        join_kind: DatasetJoinKind = DatasetJoinKind.INNER,
    ) -> None:
        match join_kind:
            case DatasetJoinKind.INNER:
                join = "INNER"
            case DatasetJoinKind.LEFT_OUTER:
                join = "LEFT"
            case DatasetJoinKind.RIGHT_OUTER:
                join = "RIGHT"
            case DatasetJoinKind.OUTER:
                join = "FULL OUTER"
            case _:
                raise NotImplementedError(f"Join kind {join_kind} is not supported.")

        # Don't quote the join column names, since they're already quoted as part of escape_identifier's output
        join_query = f"""
        CREATE TABLE {table_name} AS
        SELECT *
        FROM {table_1} a
        {join} JOIN {table_2} b
        ON a.{escape_identifier(table_1_join_key)} = b.{escape_identifier(table_2_join_key)}
        """

        conn.sql(join_query)


def _infer_dataframe_datetime_columns(df: pd.DataFrame, n: int = 100) -> list[str]:
    """
    Infer datetime columns in a pandas DataFrame by parsing non-null values in the first n rows. Return the column names believed to be datetime type

    Parameters:
        df (pandas.DataFrame): Input DataFrame.
        n (int): Number of non-null rows to consider for each column. Default is 100.

    Returns:
        datetime_columns (list): List of column names inferred to be datetime.
    """
    datetime_columns = []

    for column in df.columns:
        non_null_values = df[column].dropna().head(n)
        if non_null_values.empty:
            continue

        # Try parsing each non-null value in the column
        try:
            parsed_values = non_null_values.apply(lambda x: parse(x))

            # If parsing succeeds for all values, consider the column as datetime
            if parsed_values.notnull().all():
                datetime_columns.append(column)
        except:
            # If parsing fails for any value, move to the next column
            continue

    return datetime_columns


"""
Returns a list of ColumnFormat. Depending on structure / unstructured data, we need to format the root columns differently, so return the raw forms.

See the subtle differences between

CREATE TABLE users (
    userID BIGINT,
    userName VARCHAR,
    hobbies ARRAY<VARCHAR>
);

and

SELECT *
FROM read_json('todos.json',
               format = 'array',
               columns = {userId: 'UBIGINT',
                          userName: 'VARCHAR',
                          hobbies: 'ARRAY<VARCHAR>'});

"""


def make_duckdb_dataset_schema(schema: DatasetSchema) -> list[ColumnFormat]:
    details = []
    for col in schema.columns:
        format = _make_schema(col.definition)
        details.append(
            ColumnFormat(source_name=col.source_name, alias=str(col.id), format=format),
        )

    return details


def _make_schema(
    schema_node: DatasetObjectType | DatasetListType | DatasetScalarType,
) -> str:
    if isinstance(schema_node, DatasetObjectType):
        details = {}
        for col, value in schema_node.object.items():
            details[col] = _make_schema(value)
        key_value_pairs = [
            f"{escape_identifier(col)} {value}" for col, value in details.items()
        ]
        return f"STRUCT({', '.join(key_value_pairs)})"

    elif isinstance(schema_node, DatasetListType):
        return f"{_make_schema(schema_node.items)}[]"
    elif isinstance(schema_node, DatasetScalarType):
        match schema_node.dtype:
            case DType.INT:
                return "BIGINT"
            case DType.FLOAT:
                return "DOUBLE"
            case DType.BOOL:
                return "BOOLEAN"
            case DType.STRING | DType.IMAGE:
                return "VARCHAR"
            case DType.UUID:
                return "UUID"
            case DType.TIMESTAMP:
                return "TIMESTAMP"
            case DType.JSON:
                return "JSON"
            case _:
                raise ValueError(f"Unknown mapping for DType {schema_node.dtype}")
    else:
        raise NotImplementedError(
            f"Schema conversion not implemented for node type {type(schema_node)}",
        )


def escape_identifier(identifier: str) -> str:
    """
    Escape an identifier (e.g., column name) for use in a SQL query.
    This method handles special characters and ensures proper quoting.

    For struct fields, the identifiers must be escaped as following:
    "struct_column_name"."struct_field"
    """
    # Replace any double quotes with two double quotes
    escaped = identifier.replace('"', '""')
    # Wrap the entire identifier in double quotes and return
    return f'"{escaped}"'


def unescape_identifier(identifier: str) -> str:
    """
    Unescape an identifier (e.g., column name).

    This removes the double quotes and properly handles struct fields, which may be escaped as follows:
    "struct_column_name"."struct_field"

    Here's a hard case for help understanding this function: "struct "" column name with quotes"."struct.field.name.with.dots"
    """
    unescaped_identifiers = []
    # strip top-level quotes
    identifier = identifier[1:-1]
    # split identifier into struct fields based on delimiter pattern "."
    # at this point there are no external double quotes left; any remaining are escaped double quotes belonging to
    # the column name
    identifier_split_in_struct_fields = re.split(r'"\."', identifier)

    for identifier in identifier_split_in_struct_fields:
        # replace any escaped double quotes in the column
        unescaped_identifier = identifier.replace('""', '"')
        unescaped_identifiers.append(unescaped_identifier)

    # join back any struct fields via dot syntax without the escape identifiers
    return ".".join(unescaped_identifiers)


def escape_str_literal(literal: str) -> str:
    """
    Escape a duckDB string literal for use in a SQL query.
    https://duckdb.org/docs/stable/sql/data_types/literal_types.html#escape-string-literals
    """
    # replace any single quotes with two single quotes
    escaped = literal.replace("'", "''")
    # Wrap the entire identifier in single quotes and return
    return f"'{escaped}'"



import warnings
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from litellm.types.llms.anthropic import AnthropicThinkingParam
from pydantic import BaseModel, Field, field_validator, model_validator

warnings.filterwarnings(
    "ignore",
    message='Field name "schema".*shadows an attribute in parent "BaseModel"',
    category=UserWarning,
)


class ModelProvider(str, Enum):
    ANTHROPIC = "anthropic"
    OPENAI = "openai"
    GEMINI = "gemini"

    # not supporting for v0
    # AZURE = "azure"
    # BEDROCK = "bedrock"
    # SAGEMAKER = "sagemaker"
    # VERTEX_AI = "vertex_ai"
    # MISTRAL = "mistral"
    # META_LLAMA = "meta_llama"
    # GROQ = "groq"
    # HUGGINGFACE = "huggingface"
    # CLOUDFLARE = "cloudflare"
    # DEEPSEEK = "deepseek"
    # AI21 = "ai21"
    # BASETEN = "baseten"
    # COHERE = "cohere"
    # EMPOWER = "empower"
    # FEATHERLESS_AI = "featherless_ai"
    # FRIENDLIAI = "friendliai"
    # GALADRIEL = "galadriel"
    # NEBIUS = "nebius"
    # NLP_CLOUD = "nlp_cloud"
    # NOVITA = "novita"
    # OPENROUTER = "openrouter"
    # PETALS = "petals"
    # REPLICATE = "replicate"
    # TOGETHER_AI = "together_ai"
    # VLLM = "vllm"
    # WATSONX = "watsonx"


class MessageRole(Enum):
    DEVELOPER = "developer"
    SYSTEM = "system"
    USER = "user"
    AI = "assistant"
    TOOL = "tool"


class OpenAIMessageType(str, Enum):
    TEXT = "text"
    IMAGE_URL = "image_url"
    INPUT_AUDIO = "input_audio"


class ReasoningEffortEnum(str, Enum):
    NONE = "none"
    MINIMAL = "minimal"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    DEFAULT = "default"


class ToolChoiceEnum(str, Enum):
    AUTO = "auto"
    NONE = "none"
    REQUIRED = "required"


class LLMResponseFormatEnum(str, Enum):
    TEXT = "text"
    JSON_OBJECT = "json_object"
    JSON_SCHEMA = "json_schema"


class JsonPropertySchema(BaseModel):
    type: str = Field(
        default="string",
        description="The argument's type (e.g. string, boolean, etc.)",
    )
    description: Optional[str] = Field(
        default=None,
        description="A description of the argument",
    )
    enum: Optional[List[str]] = Field(
        default=None,
        description="An enum for the argument (e.g. ['celsius', 'fahrenheit'])",
    )
    items: Optional[Any] = Field(
        default=None,
        description="For array types, describes the items",
    )


class JsonSchema(BaseModel):
    type: str = Field(default="object")
    properties: Dict[str, JsonPropertySchema] = Field(
        description="The name of the property and the property schema (e.g. {'topic': {'type': 'string', 'description': 'the topic to generate a joke for'})",
    )
    required: List[str] = Field(
        default_factory=list,
        description="The required properties of the function",
    )
    additionalProperties: Optional[bool] = Field(
        default=None,
        description="Whether the function definition should allow additional properties",
    )


class StreamOptions(BaseModel):
    include_usage: Optional[bool] = Field(
        None,
        description="Whether to include usage information in the stream",
    )


class LogitBiasItem(BaseModel):
    token_id: int = Field(description="Token ID to bias")
    bias: float = Field(
        ge=-100,
        le=100,
        description="Bias value between -100 and 100",
    )


class ToolCallFunction(BaseModel):
    name: str = Field(description="Name of the function to call")
    arguments: str = Field(description="JSON string of function arguments")


class ToolCall(BaseModel):
    type: str = Field(
        default="function",
        description="The type of tool call. Currently the only type supported is 'function'.",
    )
    id: str = Field(description="Unique identifier for the tool call")
    function: ToolCallFunction = Field(description="Function details")

    @field_validator("type", mode="before")
    @classmethod
    def force_type(cls, v: str) -> str:
        return "function"


class ImageURL(BaseModel):
    url: str = Field(description="URL of the image")


class InputAudio(BaseModel):
    data: str = Field(description="Base64 encoded audio data")
    format: str = Field(
        description="audio format (e.g. 'mp3', 'wav', 'flac', etc.)",
    )


class OpenAIMessageItem(BaseModel):
    type: OpenAIMessageType = Field(
        description="Type of the message (either 'text', 'image_url', or 'input_audio')",
    )
    text: Optional[str] = Field(
        default=None,
        description="Text content of the message if type is 'text'",
    )
    image_url: Optional[ImageURL] = Field(
        default=None,
        description="Image URL content of the message if type is 'image_url'",
    )
    input_audio: Optional[InputAudio] = Field(
        default=None,
        description="Input audio content of the message if type is 'input_audio'",
    )

    class Config:
        use_enum_values = True


class OpenAIMessage(BaseModel):
    """
    The message schema class for the prompts playground.
    This class adheres to OpenAI's message schema.
    """

    role: MessageRole = Field(description="Role of the message")
    name: Optional[str] = Field(
        default=None,
        description="An optional name for the participant. Provides the model information to differentiate between participants of the same role.",
    )
    content: Optional[str | List[OpenAIMessageItem]] = Field(
        default=None,
        description="Content of the message",
    )
    tool_calls: Optional[List[ToolCall]] = Field(
        default=None,
        description="Tool calls made by assistant",
    )
    tool_call_id: Optional[str] = Field(
        default=None,
        description="ID of the tool call this message is responding to",
    )

    class Config:
        use_enum_values = True


class ToolFunction(BaseModel):
    name: str = Field(description="The name of the tool/function")
    description: Optional[str] = Field(
        default=None,
        description="Description of what the tool does",
    )
    parameters: Optional[JsonSchema] = Field(
        default=None,
        description="The function's parameter schema",
    )


class LLMTool(BaseModel):
    type: str = Field(
        default="function",
        description="The type of tool. Should always be 'function'",
    )
    function: ToolFunction = Field(description="The function definition")
    strict: Optional[bool] = Field(
        default=None,
        description="Whether the function definition should use OpenAI's strict mode",
    )

    @field_validator("type", mode="before")
    @classmethod
    def force_type(cls, v: str) -> str:
        return "function"


class LLMResponseSchema(BaseModel):
    name: str = Field(description="Name of the schema")
    description: Optional[str] = Field(None, description="Description of the schema")
    schema: JsonSchema = Field(description="The JSON schema object")  # type: ignore[assignment]
    strict: Optional[bool] = Field(
        None,
        description="Whether to enforce strict schema adherence",
    )


class LLMResponseFormat(BaseModel):
    type: LLMResponseFormatEnum = Field(
        description="Response format type: 'text', 'json_object', or 'json_schema'",
        examples=["json_schema"],
    )
    json_schema: Optional[LLMResponseSchema] = Field(
        None,
        description="JSON schema definition (required when type is 'json_schema')",
    )

    @model_validator(mode="after")
    def validate_schema_requirement(self) -> "LLMResponseFormat":
        if self.type == LLMResponseFormatEnum.JSON_SCHEMA and not self.json_schema:
            raise ValueError(
                "json_schema object is required when using type='json_schema'",
            )
        if (
            self.type != LLMResponseFormatEnum.JSON_SCHEMA
            and self.json_schema is not None
        ):
            raise ValueError(
                f'response format must only be {{"type": "{self.type}"}} when using type="{self.type}"',
            )
        return self

    class Config:
        use_enum_values = True
        extra = "forbid"


class ToolChoiceFunction(BaseModel):
    name: str = Field(description="The name of the function")


class ToolChoice(BaseModel):
    type: str = Field(
        default="function",
        description="The type of tool choice. Should always be 'function'",
    )
    function: Optional[ToolChoiceFunction] = Field(
        None,
        description="The tool choice fucntion name",
    )

    @field_validator("type", mode="before")
    @classmethod
    def force_type(cls, v: str) -> str:
        return "function"


class LLMBaseConfigSettings(BaseModel):
    model_config = {"extra": "forbid"}

    timeout: Optional[float] = Field(None, description="Request timeout in seconds")
    temperature: Optional[float] = Field(
        None,
        description="Sampling temperature (0.0 to 2.0). Higher values make output more random",
    )
    top_p: Optional[float] = Field(
        None,
        description="Top-p sampling parameter (0.0 to 1.0). Alternative to temperature",
    )
    max_tokens: Optional[int] = Field(
        None,
        description="Maximum number of tokens to generate in the response",
    )
    stop: Optional[str] = Field(
        None,
        description="Stop sequence(s) where the model should stop generating",
    )
    presence_penalty: Optional[float] = Field(
        None,
        description="Presence penalty (-2.0 to 2.0). Positive values penalize new tokens based on their presence",
    )
    frequency_penalty: Optional[float] = Field(
        None,
        description="Frequency penalty (-2.0 to 2.0). Positive values penalize tokens based on frequency",
    )
    seed: Optional[int] = Field(
        None,
        description="Random seed for reproducible outputs",
    )
    logprobs: Optional[bool] = Field(
        None,
        description="Whether to return log probabilities of output tokens",
    )
    top_logprobs: Optional[int] = Field(
        None,
        description="Number of most likely tokens to return log probabilities for (1-20)",
    )
    logit_bias: Optional[List[LogitBiasItem]] = Field(
        None,
        description="Modify likelihood of specified tokens appearing in completion",
    )
    max_completion_tokens: Optional[int] = Field(
        None,
        description="Maximum number of completion tokens (alternative to max_tokens)",
    )
    reasoning_effort: Optional[ReasoningEffortEnum] = Field(
        None,
        description="Reasoning effort level for models that support it (e.g., OpenAI o1 series)",
    )
    thinking: Optional[AnthropicThinkingParam] = Field(
        None,
        description="Anthropic-specific thinking parameter for Claude models",
    )


class LLMConfigSettings(LLMBaseConfigSettings):
    model_config = {"extra": "forbid"}

    tool_choice: Optional[Union[ToolChoiceEnum, ToolChoice]] = Field(
        None,
        description="Tool choice configuration ('auto', 'none', 'required', or a specific tool selection)",
    )
    response_format: Optional[Union[LLMResponseFormat, Type[BaseModel]]] = Field(
        None,
        description="Either a structured json_schema or a Pydantic model to enforce structured outputs.",
    )
    stream_options: Optional[StreamOptions] = Field(
        None,
        description="Additional streaming configuration options",
    )



import json
import logging
from datetime import datetime
from typing import Annotated, Any
from uuid import UUID

import pandas as pd
from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import (
    NumericAggregationFunction,
    SketchAggregationFunction,
)
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
    SketchMetric,
)
from arthur_common.models.schema_definitions import MetricDatasetParameterAnnotation

# Global threshold for pass/fail determination
RELEVANCE_SCORE_THRESHOLD = 0.5
TOOL_SCORE_PASS_VALUE = 1
TOOL_SCORE_NO_TOOL_VALUE = 2

logger = logging.getLogger(__name__)


def root_span_in_time_buckets(
    ddb_conn: DuckDBPyConnection, dataset: DatasetReference
) -> pd.DataFrame:
    return ddb_conn.sql(
        f"""
            SELECT
                time_bucket(INTERVAL '5 minutes', start_time) as ts,
                root_spans
            FROM {dataset.dataset_table_name}
            WHERE root_spans IS NOT NULL AND length(root_spans) > 0
            ORDER BY ts DESC;
            """,
    ).df()


def span_parser(span_to_parse: str | dict[str, Any]) -> dict[str, Any]:
    if isinstance(span_to_parse, str):
        return json.loads(span_to_parse)  # type: ignore[no-any-return]

    return span_to_parse


def extract_agent_name_from_span(span: dict[str, Any]) -> str | None:
    try:
        raw_data = span.get("raw_data", {})
        if isinstance(raw_data, str):
            raw_data = json.loads(raw_data)

        # Try to get agent name from the span's name field
        agent_name = raw_data.get("name", "unknown")
        if agent_name != "unknown":
            return str(agent_name)
    except (json.JSONDecodeError, KeyError, TypeError):
        logger.error(
            f"Error parsing attributes from span (span_id: {span.get('span_id')}) in trace {span.get('trace_id')}",
        )

    return None


# TODO: create TypedDict for span
def extract_spans_with_metrics_and_agents(
    root_spans: list[str | dict[str, Any]],
) -> list[tuple[dict[str, Any], str]]:
    """Recursively extract all spans with metrics and their associated agent names from the span tree.

    Returns:
        List of tuples: (span, agent_name)
    """
    spans_with_metrics_and_agents = []

    # TODO: Improve function so it won't modify variable outside of its scope
    def traverse_spans(
        spans: list[str | dict[str, Any]],
        current_agent: str = "unknown",
    ) -> None:
        for span_to_parse in spans:
            parsed_span = span_parser(span_to_parse)

            # Update current agent name if this span is an AGENT
            if parsed_span.get("span_kind") == "AGENT":
                agent_name = extract_agent_name_from_span(parsed_span)
                if agent_name:
                    current_agent = agent_name

            # Check if this span has metrics
            if parsed_span.get("metric_results", []):
                spans_with_metrics_and_agents.append(
                    (parsed_span, current_agent),
                )

            # Recursively traverse children with the current agent name
            if children_span := parsed_span.get("children", []):
                traverse_spans(children_span, current_agent)

    traverse_spans(root_spans)
    return spans_with_metrics_and_agents


def determine_relevance_pass_fail(score: float | None) -> str | None:
    """Determine pass/fail for relevance scores using global threshold"""
    if score is None:
        return None
    return "pass" if score >= RELEVANCE_SCORE_THRESHOLD else "fail"


def determine_tool_pass_fail(score: int | None) -> str | None:
    """Determine pass/fail for tool scores using global threshold"""
    if score is None:
        return None
    if score == TOOL_SCORE_PASS_VALUE:
        return "pass"
    elif score == TOOL_SCORE_NO_TOOL_VALUE:
        return "no_tool"
    else:
        return "fail"


class AgenticMetricsOverTimeAggregation(SketchAggregationFunction):
    """Combined aggregation for tool selection, tool usage, query relevance, and response relevance over time"""

    METRIC_NAME = "agentic_metrics_over_time"
    TOOL_SELECTION_METRIC_NAME = "tool_selection_over_time"
    TOOL_USAGE_METRIC_NAME = "tool_usage_over_time"
    QUERY_RELEVANCE_SCORES_METRIC_NAME = "query_relevance_scores_over_time"
    RESPONSE_RELEVANCE_SCORES_METRIC_NAME = "response_relevance_scores_over_time"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000030")

    @staticmethod
    def display_name() -> str:
        return "Agentic Metrics Over Time"

    @staticmethod
    def description() -> str:
        return "Metric that reports distributions (data sketches) on tool selection, tool usage, query relevance, and response relevance scores over time."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticMetricsOverTimeAggregation.TOOL_SELECTION_METRIC_NAME,
                description="Distribution of tool selection over time.",
            ),
            BaseReportedAggregation(
                metric_name=AgenticMetricsOverTimeAggregation.TOOL_USAGE_METRIC_NAME,
                description="Distribution of tool usage over time.",
            ),
            BaseReportedAggregation(
                metric_name=AgenticMetricsOverTimeAggregation.QUERY_RELEVANCE_SCORES_METRIC_NAME,
                description="Distribution of query relevance over time.",
            ),
            BaseReportedAggregation(
                metric_name=AgenticMetricsOverTimeAggregation.RESPONSE_RELEVANCE_SCORES_METRIC_NAME,
                description="Distribution of response relevance over time.",
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[SketchMetric]:
        # Query traces by timestamp
        results = root_span_in_time_buckets(ddb_conn, dataset)

        # Process traces and extract spans with metrics
        tool_selection_data = []
        tool_usage_data = []
        query_relevance_data = []
        response_relevance_data = []

        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Extract all spans with metrics and their agent names from the tree
            spans_with_metrics_and_agents = extract_spans_with_metrics_and_agents(
                root_spans,
            )

            # Process each span with metrics
            for span, agent_name in spans_with_metrics_and_agents:
                metric_results = span.get("metric_results", [])

                for metric_result in metric_results:
                    metric_type = metric_result.get("metric_type")
                    details = json.loads(metric_result.get("details", "{}"))

                    if metric_type == "ToolSelection":
                        tool_selection = details.get("tool_selection", {})

                        # Extract tool selection data
                        tool_selection_score = tool_selection.get("tool_selection")
                        tool_selection_reason = tool_selection.get(
                            "tool_selection_reason",
                            "Unknown",
                        )

                        if tool_selection_score is not None:
                            tool_selection_data.append(
                                {
                                    "ts": ts,
                                    "tool_selection_score": tool_selection_score,
                                    "tool_selection_reason": tool_selection_reason,
                                    "agent_name": agent_name,
                                },
                            )

                        # Extract tool usage data
                        tool_usage_score = tool_selection.get("tool_usage")
                        tool_usage_reason = tool_selection.get(
                            "tool_usage_reason",
                            "Unknown",
                        )

                        if tool_usage_score is not None:
                            tool_usage_data.append(
                                {
                                    "ts": ts,
                                    "tool_usage_score": tool_usage_score,
                                    "tool_usage_reason": tool_usage_reason,
                                    "agent_name": agent_name,
                                },
                            )

                    elif metric_type == "QueryRelevance":
                        query_relevance = details.get("query_relevance", {})
                        reason = query_relevance.get("reason", "Unknown")

                        # Add individual scores if they exist
                        llm_score = query_relevance.get("llm_relevance_score")
                        reranker_score = query_relevance.get("reranker_relevance_score")
                        bert_score = query_relevance.get("bert_f_score")

                        if llm_score is not None:
                            query_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "llm_relevance_score",
                                    "score_value": llm_score,
                                    "reason": reason,
                                    "agent_name": agent_name,
                                },
                            )

                        if reranker_score is not None:
                            query_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "reranker_relevance_score",
                                    "score_value": reranker_score,
                                    "reason": reason,
                                    "agent_name": agent_name,
                                },
                            )

                        if bert_score is not None:
                            query_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "bert_f_score",
                                    "score_value": bert_score,
                                    "reason": reason,
                                    "agent_name": agent_name,
                                },
                            )

                    elif metric_type == "ResponseRelevance":
                        response_relevance = details.get("response_relevance", {})
                        reason = response_relevance.get("reason", "Unknown")

                        # Add individual scores if they exist
                        llm_score = response_relevance.get("llm_relevance_score")
                        reranker_score = response_relevance.get(
                            "reranker_relevance_score",
                        )
                        bert_score = response_relevance.get("bert_f_score")

                        if llm_score is not None:
                            response_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "llm_relevance_score",
                                    "score_value": llm_score,
                                    "reason": reason,
                                    "agent_name": agent_name,
                                },
                            )

                        if reranker_score is not None:
                            response_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "reranker_relevance_score",
                                    "score_value": reranker_score,
                                    "reason": reason,
                                },
                            )

                        if bert_score is not None:
                            response_relevance_data.append(
                                {
                                    "ts": ts,
                                    "score_type": "bert_f_score",
                                    "score_value": bert_score,
                                    "reason": reason,
                                    "agent_name": agent_name,
                                },
                            )

        metrics = []

        # Create tool selection metric
        if tool_selection_data:
            df = pd.DataFrame(tool_selection_data)
            series = self.group_query_results_to_sketch_metrics(
                df,
                "tool_selection_score",
                ["tool_selection_reason", "agent_name"],
                "ts",
            )
            metrics.append(
                self.series_to_metric(self.TOOL_SELECTION_METRIC_NAME, series),
            )

        # Create tool usage metric
        if tool_usage_data:
            df = pd.DataFrame(tool_usage_data)
            series = self.group_query_results_to_sketch_metrics(
                df,
                "tool_usage_score",
                ["tool_usage_reason", "agent_name"],
                "ts",
            )
            metrics.append(self.series_to_metric(self.TOOL_USAGE_METRIC_NAME, series))

        # Create comprehensive query relevance metric (includes all score data)
        if query_relevance_data:
            df = pd.DataFrame(query_relevance_data)
            series = self.group_query_results_to_sketch_metrics(
                df,
                "score_value",
                ["score_type", "reason"],
                "ts",
            )
            metrics.append(
                self.series_to_metric(self.QUERY_RELEVANCE_SCORES_METRIC_NAME, series),
            )

        # Create comprehensive response relevance metric (includes all score data)
        if response_relevance_data:
            df = pd.DataFrame(response_relevance_data)
            series = self.group_query_results_to_sketch_metrics(
                df,
                "score_value",
                ["score_type", "reason"],
                "ts",
            )
            metrics.append(
                self.series_to_metric(
                    self.RESPONSE_RELEVANCE_SCORES_METRIC_NAME,
                    series,
                ),
            )

        return metrics


class AgenticRelevancePassFailCountAggregation(NumericAggregationFunction):
    """Combined aggregation for query and response relevance pass/fail counts by agent"""

    METRIC_NAME = "relevance_pass_fail_count"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000034")

    @staticmethod
    def display_name() -> str:
        return "Relevance Pass/Fail Count by Agent"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of query and response relevance passes and failures, segmented by agent name and metric type."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticRelevancePassFailCountAggregation.METRIC_NAME,
                description=AgenticRelevancePassFailCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[NumericMetric]:
        results = root_span_in_time_buckets(ddb_conn, dataset)

        # Process traces and extract spans with metrics
        processed_data = []
        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Extract all spans with metrics and their agent names from the tree
            spans_with_metrics_and_agents = extract_spans_with_metrics_and_agents(
                root_spans,
            )

            # Process each span with metrics
            for span, agent_name in spans_with_metrics_and_agents:
                metric_results = span.get("metric_results", [])

                for metric_result in metric_results:
                    metric_type = metric_result.get("metric_type")
                    details = json.loads(metric_result.get("details", "{}"))

                    if metric_type in ["QueryRelevance", "ResponseRelevance"]:
                        relevance_data = details.get(
                            (
                                "query_relevance"
                                if metric_type == "QueryRelevance"
                                else "response_relevance"
                            ),
                            {},
                        )
                        # Check individual scores
                        for score_type in [
                            "llm_relevance_score",
                            "reranker_relevance_score",
                            "bert_f_score",
                        ]:
                            score = relevance_data.get(score_type)
                            if score is not None:
                                result = determine_relevance_pass_fail(score)
                                processed_data.append(
                                    {
                                        "ts": ts,
                                        "agent_name": agent_name,
                                        "metric_type": metric_type,
                                        "score_type": score_type,
                                        "result": result,
                                        "count": 1,
                                    },
                                )

        if not processed_data:
            return []

        # Convert to DataFrame and aggregate
        df = pd.DataFrame(processed_data)
        aggregated = (
            df.groupby(["ts", "agent_name", "metric_type", "score_type", "result"])[
                "count"
            ]
            .sum()
            .reset_index()
        )

        series = self.group_query_results_to_numeric_metrics(
            aggregated,
            "count",
            ["agent_name", "metric_type", "score_type", "result"],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticToolPassFailCountAggregation(NumericAggregationFunction):
    """Combined aggregation for tool selection and usage pass/fail counts by agent"""

    METRIC_NAME = "tool_pass_fail_count"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000035")

    @staticmethod
    def display_name() -> str:
        return "Tool Pass/Fail Count by Agent"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of tool selection and usage passes, failures, and no-tool cases, segmented by agent name."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticToolPassFailCountAggregation.METRIC_NAME,
                description=AgenticToolPassFailCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[NumericMetric]:
        results = root_span_in_time_buckets(ddb_conn, dataset)

        # Process traces and extract spans with metrics
        processed_data = []
        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Extract all spans with metrics and their agent names from the tree
            spans_with_metrics_and_agents = extract_spans_with_metrics_and_agents(
                root_spans,
            )

            # Process each span with metrics
            for span, agent_name in spans_with_metrics_and_agents:
                metric_results = span.get("metric_results", [])

                for metric_result in metric_results:
                    if metric_result.get("metric_type") == "ToolSelection":
                        details = json.loads(metric_result.get("details", "{}"))
                        tool_selection = details.get("tool_selection", {})

                        tool_selection_score = tool_selection.get("tool_selection")
                        tool_usage_score = tool_selection.get("tool_usage")

                        # Process tool selection
                        if tool_selection_score is not None:
                            result = determine_tool_pass_fail(tool_selection_score)
                            processed_data.append(
                                {
                                    "ts": ts,
                                    "agent_name": agent_name,
                                    "tool_metric": "tool_selection",
                                    "result": result,
                                    "count": 1,
                                },
                            )

                        # Process tool usage
                        if tool_usage_score is not None:
                            result = determine_tool_pass_fail(tool_usage_score)
                            processed_data.append(
                                {
                                    "ts": ts,
                                    "agent_name": agent_name,
                                    "tool_metric": "tool_usage",
                                    "result": result,
                                    "count": 1,
                                },
                            )

        if not processed_data:
            return []

        # Convert to DataFrame and aggregate
        df = pd.DataFrame(processed_data)
        aggregated = (
            df.groupby(["ts", "agent_name", "tool_metric", "result"])["count"]
            .sum()
            .reset_index()
        )

        series = self.group_query_results_to_numeric_metrics(
            aggregated,
            "count",
            ["agent_name", "tool_metric", "result"],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticEventCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "event_count"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000036")

    @staticmethod
    def display_name() -> str:
        return "Number of Events"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of events over time."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticEventCountAggregation.METRIC_NAME,
                description=AgenticEventCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[NumericMetric]:
        results = ddb_conn.sql(
            f"""
            SELECT
                time_bucket(INTERVAL '5 minutes', start_time) as ts,
                COUNT(*) as count
            FROM {dataset.dataset_table_name}
            GROUP BY ts
            ORDER BY ts DESC;
            """,
        ).df()

        series = self.group_query_results_to_numeric_metrics(
            results,
            "count",
            [],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticLLMCallCountAggregation(NumericAggregationFunction):
    METRIC_NAME = "llm_call_count"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000038")

    @staticmethod
    def display_name() -> str:
        return "Number of LLM Calls"

    @staticmethod
    def description() -> str:
        return "Metric that counts the number of LLM spans (individual LLM calls) over time."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticLLMCallCountAggregation.METRIC_NAME,
                description=AgenticLLMCallCountAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[NumericMetric]:
        results = root_span_in_time_buckets(ddb_conn, dataset)

        # Process traces and count LLM spans
        llm_call_counts = {}
        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Count LLM spans in the tree
            def count_llm_spans(spans: list[str | dict[str, Any]]) -> int:
                count = 0
                for span_to_parse in spans:
                    span = span_parser(span_to_parse)

                    # Check if this span is an LLM span
                    if span.get("span_kind") == "LLM":
                        count += 1

                    # Recursively count children
                    if span.get("children"):
                        count += count_llm_spans(span["children"])
                return count

            llm_count = count_llm_spans(root_spans)

            if llm_count > 0:
                if ts not in llm_call_counts:
                    llm_call_counts[ts] = 0
                llm_call_counts[ts] += llm_count

        if not llm_call_counts:
            return []

        # Convert to DataFrame format
        data = [{"ts": ts, "count": count} for ts, count in llm_call_counts.items()]
        df = pd.DataFrame(data)

        series = self.group_query_results_to_numeric_metrics(
            df,
            "count",
            [],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticToolSelectionAndUsageByAgentAggregation(NumericAggregationFunction):
    METRIC_NAME = "tool_selection_and_usage_by_agent"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000037")

    @staticmethod
    def display_name() -> str:
        return "Tool Selection and Usage by Agent"

    @staticmethod
    def description() -> str:
        return "Metric that counts tool selection and usage correctness, segmented by agent name."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticToolSelectionAndUsageByAgentAggregation.METRIC_NAME,
                description=AgenticToolSelectionAndUsageByAgentAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[NumericMetric]:
        # Query traces by timestamp
        results = root_span_in_time_buckets(ddb_conn, dataset)

        # Process traces and extract spans with metrics
        processed_data = []
        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Extract all spans with metrics and their agent names from the tree
            spans_with_metrics_and_agents = extract_spans_with_metrics_and_agents(
                root_spans,
            )

            # Process each span with metrics
            for span, agent_name in spans_with_metrics_and_agents:
                metric_results = span.get("metric_results", [])

                for metric_result in metric_results:
                    if metric_result.get("metric_type") == "ToolSelection":
                        details = json.loads(metric_result.get("details", "{}"))
                        tool_selection = details.get("tool_selection", {})

                        tool_selection_score = tool_selection.get("tool_selection")
                        tool_usage_score = tool_selection.get("tool_usage")

                        if tool_selection_score is not None:
                            # Categorize selection
                            if tool_selection_score == 1:
                                selection_category = "correct_selection"
                            elif tool_selection_score == 0:
                                selection_category = "incorrect_selection"
                            else:
                                selection_category = "no_selection"

                            # Categorize usage
                            if tool_usage_score == 1:
                                usage_category = "correct_usage"
                            elif tool_usage_score == 0:
                                usage_category = "incorrect_usage"
                            else:
                                usage_category = "no_usage"

                            processed_data.append(
                                {
                                    "ts": ts,
                                    "agent_name": agent_name,
                                    "selection_category": selection_category,
                                    "usage_category": usage_category,
                                    "count": 1,
                                },
                            )

        if not processed_data:
            return []

        # Convert to DataFrame and aggregate
        df = pd.DataFrame(processed_data)
        aggregated = (
            df.groupby(["ts", "agent_name", "selection_category", "usage_category"])[
                "count"
            ]
            .sum()
            .reset_index()
        )

        series = self.group_query_results_to_numeric_metrics(
            aggregated,
            "count",
            ["agent_name", "selection_category", "usage_category"],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticTraceLatencyAggregation(SketchAggregationFunction):
    METRIC_NAME = "trace_latency"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000039")

    @staticmethod
    def display_name() -> str:
        return "Trace Latency"

    @staticmethod
    def description() -> str:
        return "Aggregation that reports the latency of the agentic trace in ms."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticTraceLatencyAggregation.METRIC_NAME,
                description=AgenticTraceLatencyAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[SketchMetric]:
        # Query traces by timestamp and calculate latency directly in SQL
        results = ddb_conn.sql(
            f"""
            SELECT
                time_bucket(INTERVAL '5 minutes', start_time) as ts,
                CAST(EXTRACT(EPOCH FROM (end_time - start_time)) * 1000 AS INTEGER) as latency_ms
            FROM {dataset.dataset_table_name}
            WHERE start_time IS NOT NULL 
                AND end_time IS NOT NULL
                AND end_time > start_time
            ORDER BY ts DESC;
            """,
        ).df()

        if results.empty:
            return []

        df = results
        # Create a single time series without grouping dimensions
        # Since we have no dimensions to group by, we create one time series for all data
        series = [self._group_to_series(df, "ts", [], "latency_ms")]
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]


class AgenticSpanLatencyAggregation(SketchAggregationFunction):
    METRIC_NAME = "span_latency"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-000000000040")

    @staticmethod
    def display_name() -> str:
        return "Span Latency"

    @staticmethod
    def description() -> str:
        return "Aggregation that reports the latency of the agentic span in ms."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=AgenticSpanLatencyAggregation.METRIC_NAME,
                description=AgenticSpanLatencyAggregation.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The agentic trace dataset containing traces with nested spans.",
                model_problem_type=ModelProblemType.AGENTIC_TRACE,
            ),
        ],
    ) -> list[SketchMetric]:
        results = root_span_in_time_buckets(ddb_conn, dataset)

        latency_data = []
        for _, row in results.iterrows():
            ts = row["ts"]
            root_spans = row["root_spans"]

            # Parse root_spans if it's a string
            if isinstance(root_spans, str):
                root_spans = json.loads(root_spans)

            # Extract all spans with their timing data
            spans_with_timing = self._extract_spans_with_timing(root_spans)

            for span_data in spans_with_timing:
                span, current_agent, latency_ms = span_data
                span_kind = span.get("span_kind", "unknown")

                if latency_ms is not None and latency_ms > 0:
                    latency_data.append(
                        {
                            "ts": ts,
                            "latency_ms": latency_ms,
                            "span_kind": span_kind,
                            "agent_name": current_agent,
                        }
                    )

        if not latency_data:
            return []

        # Convert to DataFrame and create sketch metrics
        df = pd.DataFrame(latency_data)
        series = self.group_query_results_to_sketch_metrics(
            df,
            "latency_ms",
            ["span_kind", "agent_name"],
            "ts",
        )
        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]

    def _extract_spans_with_timing(
        self, spans: list[str | dict[str, Any]], current_agent: str = "unknown"
    ) -> list[tuple[dict[str, Any], str, int | None]]:
        """Recursively extract spans with calculated latency in milliseconds"""
        spans_with_timing = []

        for span_to_parse in spans:
            span = span_parser(span_to_parse)

            # Update current agent name if this span is an AGENT
            if span.get("span_kind") == "AGENT":
                agent_name = extract_agent_name_from_span(span)
                if agent_name:
                    current_agent = agent_name

            # Calculate latency if both start_time and end_time exist
            start_time = span.get("start_time")
            end_time = span.get("end_time")
            latency_ms = None

            if start_time and end_time:
                try:
                    # Parse ISO format timestamps and calculate latency in milliseconds
                    # Assume same timezone for start and end time, specific TZ not important for latency calculation
                    start_dt = datetime.fromisoformat(start_time)
                    end_dt = datetime.fromisoformat(end_time)
                    latency_ms = int((end_dt - start_dt).total_seconds() * 1000)
                except (ValueError, TypeError) as e:
                    logger.warning(
                        f"Error calculating latency for span {span.get('span_id')}: {e}"
                    )

            spans_with_timing.append((span, current_agent, latency_ms))

            # Recursively process children
            if children := span.get("children", []):
                spans_with_timing.extend(
                    self._extract_spans_with_timing(children, current_agent)
                )

        return spans_with_timing



from __future__ import annotations

from enum import Enum
from typing import Optional, Self, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, computed_field, model_validator

from arthur_common.models.enums import ModelProblemType


class ScopeSchemaTag(str, Enum):
    LLM_CONTEXT = "llm_context"
    LLM_PROMPT = "llm_prompt"
    LLM_RESPONSE = "llm_response"
    PRIMARY_TIMESTAMP = "primary_timestamp"
    CATEGORICAL = "categorical"
    CONTINUOUS = "continuous"
    PREDICTION = "prediction"
    GROUND_TRUTH = "ground_truth"
    PIN_IN_DEEP_DIVE = "pin_in_deep_dive"
    POSSIBLE_SEGMENTATION = "possible_segmentation"


class DType(str, Enum):
    UNDEFINED = "undefined"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    STRING = "str"
    UUID = "uuid"
    TIMESTAMP = "timestamp"
    DATE = "date"
    JSON = "json"
    IMAGE = "image"


class MetricParameterAnnotation(BaseModel):
    optional: bool = Field(
        False,
        description="Boolean denoting if the parameter is optional.",
    )
    friendly_name: str = Field(
        description="User facing name of the parameter.",
    )
    description: str = Field(
        description="Description of the parameter.",
    )


class MetricDatasetParameterAnnotation(MetricParameterAnnotation):
    model_problem_type: Optional[ModelProblemType] = Field(
        default=None,
        description="Model problem type that is applicable to this parameter.",
    )


class MetricLiteralParameterAnnotation(MetricParameterAnnotation):
    parameter_dtype: DType = Field(description="Data type of the parameter.")


class MetricColumnParameterAnnotation(MetricParameterAnnotation):
    tag_hints: list[ScopeSchemaTag] = Field(
        [],
        description="List of tags that are applicable to this parameter. Datasets with columns that have matching tags can be inferred this way.",
    )
    source_dataset_parameter_key: str = Field(
        description="Name of the parameter that provides the dataset to be used for this column.",
    )
    allowed_column_types: Optional[list[SchemaTypeUnion]] = Field(
        default=None,
        description="List of column types applicable to this parameter",
    )
    allow_any_column_type: bool = Field(
        False,
        description="Indicates if this metric parameter can accept any column type.",
    )

    @model_validator(mode="after")
    def column_type_combination_validator(self) -> Self:
        if self.allowed_column_types and self.allow_any_column_type:
            raise ValueError(
                "Parameter cannot allow any column while also explicitly listing applicable ones.",
            )
        return self


class MetricMultipleColumnParameterAnnotation(MetricColumnParameterAnnotation):
    pass


MetricsParameterAnnotationUnion = (
    MetricDatasetParameterAnnotation
    | MetricLiteralParameterAnnotation
    | MetricColumnParameterAnnotation
    | MetricMultipleColumnParameterAnnotation
)


class Type(BaseModel):
    # There's bound to be something in common here eventually
    pass


class ScalarType(Type):
    dtype: DType

    def __hash__(self) -> int:
        return hash(self.dtype)


class ObjectType(Type):
    object: dict[str, SchemaTypeUnion]

    def __getitem__(self, key: str) -> SchemaTypeUnion:
        return self.object[key]

    def __hash__(self) -> int:
        # Combine the hash of all dictionary values
        combined_hash = 0
        for name, col in self.object.items():
            combined_hash ^= hash(name)
            combined_hash ^= hash(col)
        return combined_hash


class ListType(Type):
    items: SchemaTypeUnion

    def __hash__(self) -> int:
        return hash(self.items)


class DatasetSchemaType(Type):
    tag_hints: list[ScopeSchemaTag] = []
    nullable: Optional[bool] = True
    id: UUID = Field(default_factory=uuid4, description="Unique ID of the schema node.")


class DatasetScalarType(ScalarType, DatasetSchemaType):
    def __hash__(self) -> int:
        return hash(self.dtype)

    def to_base_type(self) -> ScalarType:
        return ScalarType(dtype=self.dtype)


class DatasetObjectType(DatasetSchemaType):
    object: dict[str, DatasetSchemaTypeUnion]

    def __getitem__(self, key: str) -> DatasetSchemaTypeUnion:
        return self.object[key]

    def __hash__(self) -> int:
        # Combine the hash of all dictionary values
        combined_hash = 0
        for name, col in self.object.items():
            combined_hash ^= hash(name)
            combined_hash ^= hash(col)
        return combined_hash

    def to_base_type(self) -> ObjectType:
        return ObjectType(object={k: v.to_base_type() for k, v in self.object.items()})


class DatasetListType(DatasetSchemaType):
    items: DatasetSchemaTypeUnion

    def __hash__(self) -> int:
        return hash(self.items)

    def to_base_type(self) -> ListType:
        return ListType(items=self.items.to_base_type())


class DatasetColumn(BaseModel):
    id: UUID = Field(default_factory=uuid4, description="Unique ID of the column.")
    source_name: str
    definition: DatasetSchemaTypeUnion

    def __hash__(self) -> int:
        combined_hash = 0
        combined_hash ^= hash(self.source_name)
        combined_hash ^= hash(self.definition)
        return combined_hash


class PutDatasetSchema(BaseModel):
    alias_mask: dict[UUID, str]
    columns: list[DatasetColumn]

    def regenerate_ids(self) -> PutDatasetSchema:
        new_columns = []
        new_alias_mask = {}
        for column in self.columns:
            new_id = uuid4()
            new_columns.append(
                DatasetColumn(
                    id=new_id,
                    source_name=column.source_name,
                    definition=self._regenerate_definition_ids(column.definition),
                ),
            )
            if column.id in self.alias_mask:
                new_alias_mask[new_id] = self.alias_mask[column.id]

        self.columns = new_columns
        self.alias_mask = new_alias_mask
        return self

    @staticmethod
    def _regenerate_definition_ids(
        definition: DatasetSchemaTypeUnion,
    ) -> DatasetSchemaTypeUnion:
        new_def = definition.model_copy(deep=True)
        new_def.id = uuid4()

        if isinstance(new_def, DatasetObjectType):
            new_def.object = {
                k: PutDatasetSchema._regenerate_definition_ids(v)
                for k, v in new_def.object.items()
            }
        elif isinstance(new_def, DatasetListType):
            new_def.items = PutDatasetSchema._regenerate_definition_ids(new_def.items)

        return new_def


# This needs to be a separate model than PutDatasetSchema because of this generated field not being consumed correctly by the client generator
# Issue tracked externally here: https://github.com/OpenAPITools/openapi-generator/issues/4190
class DatasetSchema(PutDatasetSchema):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def column_names(self) -> dict[UUID, str]:
        col_names = {column.id: column.source_name for column in self.columns}
        col_names.update(self.alias_mask)
        return col_names

    model_config = ConfigDict(
        json_schema_mode_override="serialization",
    )


SchemaTypeUnion = Union[ScalarType, ObjectType, ListType]
DatasetSchemaTypeUnion = Union[DatasetScalarType, DatasetObjectType, DatasetListType]

from uuid import uuid4


def create_dataset_scalar_type(dtype: DType) -> DatasetScalarType:
    return DatasetScalarType(id=uuid4(), dtype=dtype)


def create_dataset_object_type(
    object_dict: dict[str, DatasetSchemaTypeUnion],
) -> DatasetObjectType:
    return DatasetObjectType(id=uuid4(), object={k: v for k, v in object_dict.items()})


def create_dataset_list_type(items: DatasetSchemaTypeUnion) -> DatasetListType:
    return DatasetListType(id=uuid4(), items=items)


def create_shield_rule_results_schema() -> DatasetListType:
    return create_dataset_list_type(
        create_dataset_object_type(
            {
                "id": create_dataset_scalar_type(DType.UUID),
                "name": create_dataset_scalar_type(DType.STRING),
                "rule_type": create_dataset_scalar_type(DType.STRING),
                "scope": create_dataset_scalar_type(DType.STRING),
                "result": create_dataset_scalar_type(DType.STRING),
                "latency_ms": create_dataset_scalar_type(DType.INT),
                "details": create_dataset_object_type(
                    {
                        "message": create_dataset_scalar_type(DType.STRING),
                        "claims": create_dataset_list_type(
                            create_dataset_object_type(
                                {
                                    "claim": create_dataset_scalar_type(DType.STRING),
                                    "valid": create_dataset_scalar_type(DType.BOOL),
                                    "reason": create_dataset_scalar_type(DType.STRING),
                                },
                            ),
                        ),
                        "pii_entities": create_dataset_list_type(
                            create_dataset_object_type(
                                {
                                    "entity": create_dataset_scalar_type(DType.STRING),
                                    "span": create_dataset_scalar_type(DType.STRING),
                                    "confidence": create_dataset_scalar_type(
                                        DType.FLOAT,
                                    ),
                                },
                            ),
                        ),
                        "toxicity_score": create_dataset_scalar_type(DType.FLOAT),
                        "regex_matches": create_dataset_list_type(
                            create_dataset_object_type(
                                {
                                    "matching_text": create_dataset_scalar_type(
                                        DType.STRING,
                                    ),
                                    "pattern": create_dataset_scalar_type(DType.STRING),
                                },
                            ),
                        ),
                        "keyword_matches": create_dataset_list_type(
                            create_dataset_object_type(
                                {
                                    "keyword": create_dataset_scalar_type(DType.STRING),
                                },
                            ),
                        ),
                    },
                ),
            },
        ),
    )


def create_shield_prompt_schema() -> DatasetObjectType:
    return create_dataset_object_type(
        {
            "id": create_dataset_scalar_type(DType.UUID),
            "inference_id": create_dataset_scalar_type(DType.UUID),
            "result": create_dataset_scalar_type(DType.STRING),
            "created_at": create_dataset_scalar_type(DType.INT),
            "updated_at": create_dataset_scalar_type(DType.INT),
            "message": create_dataset_scalar_type(DType.STRING),
            "prompt_rule_results": create_shield_rule_results_schema(),
            "tokens": create_dataset_scalar_type(DType.INT),
        },
    )


def create_shield_response_schema() -> DatasetObjectType:
    return create_dataset_object_type(
        {
            "id": create_dataset_scalar_type(DType.UUID),
            "inference_id": create_dataset_scalar_type(DType.UUID),
            "result": create_dataset_scalar_type(DType.STRING),
            "created_at": create_dataset_scalar_type(DType.INT),
            "updated_at": create_dataset_scalar_type(DType.INT),
            "message": create_dataset_scalar_type(DType.STRING),
            "context": create_dataset_scalar_type(DType.STRING),
            "response_rule_results": create_shield_rule_results_schema(),
            "tokens": create_dataset_scalar_type(DType.INT),
        },
    )


def create_shield_inference_feedback_schema() -> DatasetListType:
    return create_dataset_list_type(
        create_dataset_object_type(
            {
                "id": create_dataset_scalar_type(DType.UUID),
                "inference_id": create_dataset_scalar_type(DType.UUID),
                "target": create_dataset_scalar_type(DType.STRING),
                "score": create_dataset_scalar_type(DType.FLOAT),
                "reason": create_dataset_scalar_type(DType.STRING),
                "user_id": create_dataset_scalar_type(DType.UUID),
                "created_at": create_dataset_scalar_type(DType.TIMESTAMP),
                "updated_at": create_dataset_scalar_type(DType.TIMESTAMP),
            },
        ),
    )


def AGENTIC_TRACE_SCHEMA() -> DatasetSchema:
    return DatasetSchema(
        alias_mask={},
        columns=[
            DatasetColumn(
                id=uuid4(),
                source_name="trace_id",
                definition=create_dataset_scalar_type(DType.STRING),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="start_time",
                definition=create_dataset_scalar_type(DType.TIMESTAMP),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="end_time",
                definition=create_dataset_scalar_type(DType.TIMESTAMP),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="root_spans",
                definition=create_dataset_list_type(
                    create_dataset_scalar_type(
                        DType.JSON,
                    ),  # JSON blob to preserve hierarchy
                ),
            ),
        ],
    )


def SHIELD_SCHEMA() -> DatasetSchema:
    return DatasetSchema(
        alias_mask={},
        columns=[
            DatasetColumn(
                id=uuid4(),
                source_name="id",
                definition=create_dataset_scalar_type(DType.UUID),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="result",
                definition=create_dataset_scalar_type(DType.STRING),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="created_at",
                definition=create_dataset_scalar_type(DType.INT),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="updated_at",
                definition=create_dataset_scalar_type(DType.INT),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="task_id",
                definition=create_dataset_scalar_type(DType.UUID),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="conversation_id",
                definition=create_dataset_scalar_type(DType.STRING),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="user_id",
                definition=create_dataset_scalar_type(DType.STRING),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="inference_prompt",
                definition=create_shield_prompt_schema(),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="inference_response",
                definition=create_shield_response_schema(),
            ),
            DatasetColumn(
                id=uuid4(),
                source_name="inference_feedback",
                definition=create_shield_inference_feedback_schema(),
            ),
        ],
    )


SHIELD_RESPONSE_SCHEMA = create_shield_response_schema().to_base_type()
SHIELD_PROMPT_SCHEMA = create_shield_prompt_schema().to_base_type()


# Agentic trace schema base type for API responses
def create_agentic_trace_response_schema() -> DatasetObjectType:
    return create_dataset_object_type(
        {
            "count": create_dataset_scalar_type(DType.INT),
            "traces": create_dataset_list_type(
                create_dataset_object_type(
                    {
                        "trace_id": create_dataset_scalar_type(DType.STRING),
                        "start_time": create_dataset_scalar_type(DType.TIMESTAMP),
                        "end_time": create_dataset_scalar_type(DType.TIMESTAMP),
                        "root_spans": create_dataset_list_type(
                            create_dataset_scalar_type(
                                DType.JSON,
                            ),  # JSON blob for infinite depth
                        ),
                    },
                ),
            ),
        },
    )


AGENTIC_TRACE_RESPONSE_SCHEMA = create_agentic_trace_response_schema().to_base_type()

SEGMENTATION_ALLOWED_DTYPES = [DType.INT, DType.BOOL, DType.STRING, DType.UUID]
SEGMENTATION_ALLOWED_COLUMN_TYPES = [
    ScalarType(dtype=d_type) for d_type in SEGMENTATION_ALLOWED_DTYPES
]



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.enums import ModelProblemType
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricLiteralParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)
from arthur_common.tools.duckdb_data_loader import escape_str_literal


class MulticlassClassifierStringLabelSingleClassConfusionMatrixAggregationFunction(
    NumericAggregationFunction,
):
    MULTICLASS_CM_SINGLE_CLASS_TP_COUNT_METRIC_NAME = (
        "multiclass_confusion_matrix_single_class_true_positive_count"
    )
    MULTICLASS_CM_SINGLE_CLASS_FP_COUNT_METRIC_NAME = (
        "multiclass_confusion_matrix_single_class_false_positive_count"
    )
    MULTICLASS_CM_SINGLE_CLASS_FN_COUNT_METRIC_NAME = (
        "multiclass_confusion_matrix_single_class_false_negative_count"
    )
    MULTICLASS_CM_SINGLE_CLASS_TN_COUNT_METRIC_NAME = (
        "multiclass_confusion_matrix_single_class_true_negative_count"
    )

    @staticmethod
    def id() -> UUID:
        return UUID("dc728927-6928-4a3b-b174-8c1ec8b58d62")

    @staticmethod
    def display_name() -> str:
        return "Multiclass Classification Confusion Matrix Single Class - String Class Label Prediction"

    @staticmethod
    def description() -> str:
        return (
            "Aggregation that takes in the string label for the positive class, "
            "and calculates the confusion matrix (True Positives, False Positives, "
            "False Negatives, True Negatives) for that class compared to all others."
        )

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=MulticlassClassifierStringLabelSingleClassConfusionMatrixAggregationFunction.MULTICLASS_CM_SINGLE_CLASS_TP_COUNT_METRIC_NAME,
                description="Confusion matrix true positives count.",
            ),
            BaseReportedAggregation(
                metric_name=MulticlassClassifierStringLabelSingleClassConfusionMatrixAggregationFunction.MULTICLASS_CM_SINGLE_CLASS_FP_COUNT_METRIC_NAME,
                description="Confusion matrix false positives count.",
            ),
            BaseReportedAggregation(
                metric_name=MulticlassClassifierStringLabelSingleClassConfusionMatrixAggregationFunction.MULTICLASS_CM_SINGLE_CLASS_FN_COUNT_METRIC_NAME,
                description="Confusion matrix false negatives count.",
            ),
            BaseReportedAggregation(
                metric_name=MulticlassClassifierStringLabelSingleClassConfusionMatrixAggregationFunction.MULTICLASS_CM_SINGLE_CLASS_TN_COUNT_METRIC_NAME,
                description="Confusion matrix true negatives count.",
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the prediction and ground truth values.",
                model_problem_type=ModelProblemType.MULTICLASS_CLASSIFICATION,
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        prediction_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.STRING),
                ],
                tag_hints=[ScopeSchemaTag.PREDICTION],
                friendly_name="Prediction Column",
                description="A column containing the predicted string class label.",
            ),
        ],
        gt_values_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.STRING),
                ],
                tag_hints=[ScopeSchemaTag.GROUND_TRUTH],
                friendly_name="Ground Truth Column",
                description="A column containing the ground truth string class label.",
            ),
        ],
        positive_class_label: Annotated[
            str,
            MetricLiteralParameterAnnotation(
                parameter_dtype=DType.STRING,
                friendly_name="Positive Class Label",
                description="The label indicating a positive class.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        segmentation_cols = [] if not segmentation_cols else segmentation_cols
        escaped_positive_class_label = escape_str_literal(positive_class_label)
        normalization_case = f"""
                CASE
                    WHEN value = {escaped_positive_class_label} THEN 1
                    ELSE 0
                END
                """
        return self.generate_confusion_matrix_metrics(
            ddb_conn,
            timestamp_col,
            prediction_col,
            gt_values_col,
            normalization_case,
            normalization_case,
            dataset,
            escaped_positive_class_label,
            segmentation_cols,
        )

    def generate_confusion_matrix_metrics(
        self,
        ddb_conn: DuckDBPyConnection,
        timestamp_col: str,
        prediction_col: str,
        gt_values_col: str,
        prediction_normalization_case: str,
        gt_normalization_case: str,
        dataset: DatasetReference,
        escaped_positive_class_label: str,
        segmentation_cols: list[str],
    ) -> list[NumericMetric]:
        """
        Generate a SQL query to compute confusion matrix metrics over time.

        Args:
            ddb_conn: duck DB connection
            timestamp_col: Column name containing timestamps
            prediction_col: Column name containing predictions
            gt_values_col: Column name containing ground truth values
            prediction_normalization_case: SQL CASE statement for normalizing predictions to 0 / 1 / null using 'value' as the target column name
            gt_normalization_case: SQL CASE statement for normalizing ground truth values to 0 / 1 / null using 'value' as the target column name
            dataset: DatasetReference containing dataset metadata
            escaped_positive_class_label: escaped label for the class to include in the dimensions
            segmentation_cols: List of columns to segment by

        Returns:
            str: SQL query that computes confusion matrix metrics
            Returns the following SQL with no segmentation:
            WITH normalized_data AS (
                    SELECT
                        {timestamp_col} AS timestamp,
                        {prediction_normalization_case.replace('value', prediction_col)} AS prediction,
                        {gt_normalization_case.replace('value', gt_values_col)} AS actual_value
                    FROM {dataset.dataset_table_name}
                    WHERE {timestamp_col} IS NOT NULL
                )
                SELECT
                    time_bucket(INTERVAL '5 minutes', timestamp) AS ts,
                    SUM(CASE WHEN prediction = 1 AND actual_value = 1 THEN 1 ELSE 0 END) AS true_positive_count,
                    SUM(CASE WHEN prediction = 1 AND actual_value = 0 THEN 1 ELSE 0 END) AS false_positive_count,
                    SUM(CASE WHEN prediction = 0 AND actual_value = 1 THEN 1 ELSE 0 END) AS false_negative_count,
                    SUM(CASE WHEN prediction = 0 AND actual_value = 0 THEN 1 ELSE 0 END) AS true_negative_count,
                    any_value({escaped_positive_class_label}) as class_label
                FROM normalized_data
                GROUP BY ts
                ORDER BY ts

        """
        # build query components with segmentation columns
        first_subquery_select_cols = [
            f"{timestamp_col} AS timestamp",
            f"{prediction_normalization_case.replace('value', prediction_col)} AS prediction",
            f"{gt_normalization_case.replace('value', gt_values_col)} AS actual_value",
        ] + segmentation_cols
        second_subquery_select_cols = [
            "time_bucket(INTERVAL '5 minutes', timestamp) AS ts",
            "SUM(CASE WHEN prediction = 1 AND actual_value = 1 THEN 1 ELSE 0 END) AS true_positive_count",
            "SUM(CASE WHEN prediction = 1 AND actual_value = 0 THEN 1 ELSE 0 END) AS false_positive_count",
            "SUM(CASE WHEN prediction = 0 AND actual_value = 1 THEN 1 ELSE 0 END) AS false_negative_count",
            "SUM(CASE WHEN prediction = 0 AND actual_value = 0 THEN 1 ELSE 0 END) AS true_negative_count",
            f"any_value({escaped_positive_class_label}) as class_label",
        ] + segmentation_cols
        second_subquery_group_by_cols = ["ts"] + segmentation_cols
        extra_dims = ["class_label"]

        # build query
        confusion_matrix_query = f"""
        WITH normalized_data AS (
            SELECT {", ".join(first_subquery_select_cols)}
            FROM {dataset.dataset_table_name}
            WHERE {timestamp_col} IS NOT NULL
        )
        SELECT {", ".join(second_subquery_select_cols)}
        FROM normalized_data
        GROUP BY {", ".join(second_subquery_group_by_cols)}
        ORDER BY ts
"""

        results = ddb_conn.sql(confusion_matrix_query).df()

        tp = self.group_query_results_to_numeric_metrics(
            results,
            "true_positive_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        fp = self.group_query_results_to_numeric_metrics(
            results,
            "false_positive_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        fn = self.group_query_results_to_numeric_metrics(
            results,
            "false_negative_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        tn = self.group_query_results_to_numeric_metrics(
            results,
            "true_negative_count",
            dim_columns=segmentation_cols + extra_dims,
            timestamp_col="ts",
        )
        tp_metric = self.series_to_metric(
            self.MULTICLASS_CM_SINGLE_CLASS_TP_COUNT_METRIC_NAME,
            tp,
        )
        fp_metric = self.series_to_metric(
            self.MULTICLASS_CM_SINGLE_CLASS_FP_COUNT_METRIC_NAME,
            fp,
        )
        fn_metric = self.series_to_metric(
            self.MULTICLASS_CM_SINGLE_CLASS_FN_COUNT_METRIC_NAME,
            fn,
        )
        tn_metric = self.series_to_metric(
            self.MULTICLASS_CM_SINGLE_CLASS_TN_COUNT_METRIC_NAME,
            tn,
        )
        return [tp_metric, fp_metric, fn_metric, tn_metric]



from typing import Annotated, Optional
from uuid import UUID

from duckdb import DuckDBPyConnection

from arthur_common.aggregations.aggregator import NumericAggregationFunction
from arthur_common.models.metrics import (
    BaseReportedAggregation,
    DatasetReference,
    Dimension,
    NumericMetric,
)
from arthur_common.models.schema_definitions import (
    SEGMENTATION_ALLOWED_COLUMN_TYPES,
    DType,
    MetricColumnParameterAnnotation,
    MetricDatasetParameterAnnotation,
    MetricMultipleColumnParameterAnnotation,
    ScalarType,
    ScopeSchemaTag,
)
from arthur_common.tools.duckdb_data_loader import unescape_identifier


class NumericSumAggregationFunction(NumericAggregationFunction):
    METRIC_NAME = "numeric_sum"

    @staticmethod
    def id() -> UUID:
        return UUID("00000000-0000-0000-0000-00000000000f")

    @staticmethod
    def display_name() -> str:
        return "Numeric Sum"

    @staticmethod
    def description() -> str:
        return "Metric that reports the sum of the numeric column per time window."

    @staticmethod
    def reported_aggregations() -> list[BaseReportedAggregation]:
        return [
            BaseReportedAggregation(
                metric_name=NumericSumAggregationFunction.METRIC_NAME,
                description=NumericSumAggregationFunction.description(),
            ),
        ]

    def aggregate(
        self,
        ddb_conn: DuckDBPyConnection,
        dataset: Annotated[
            DatasetReference,
            MetricDatasetParameterAnnotation(
                friendly_name="Dataset",
                description="The dataset containing the numeric data.",
            ),
        ],
        timestamp_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.TIMESTAMP),
                ],
                tag_hints=[ScopeSchemaTag.PRIMARY_TIMESTAMP],
                friendly_name="Timestamp Column",
                description="A column containing timestamp values to bucket by.",
            ),
        ],
        numeric_col: Annotated[
            str,
            MetricColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=[
                    ScalarType(dtype=DType.INT),
                    ScalarType(dtype=DType.FLOAT),
                ],
                tag_hints=[ScopeSchemaTag.CONTINUOUS],
                friendly_name="Numeric Column",
                description="A column containing numeric values to sum.",
            ),
        ],
        segmentation_cols: Annotated[
            Optional[list[str]],
            MetricMultipleColumnParameterAnnotation(
                source_dataset_parameter_key="dataset",
                allowed_column_types=SEGMENTATION_ALLOWED_COLUMN_TYPES,
                tag_hints=[ScopeSchemaTag.POSSIBLE_SEGMENTATION],
                friendly_name="Segmentation Columns",
                description="All columns to include as dimensions for segmentation.",
                optional=True,
            ),
        ] = None,
    ) -> list[NumericMetric]:
        """Executed SQL with no segmentation columns:
                select time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts, \
                sum({numeric_col}) as sum \
                from {dataset.dataset_table_name} \
                where {numeric_col} is not null \
                group by ts \
        """
        segmentation_cols = [] if not segmentation_cols else segmentation_cols

        # build query components with segmentation columns
        all_select_clause_cols = [
            f"time_bucket(INTERVAL '5 minutes', {timestamp_col}) as ts",
            f"sum({numeric_col}) as sum",
        ] + segmentation_cols
        all_group_by_cols = ["ts"] + segmentation_cols

        # build query
        query = f"""
                    select {", ".join(all_select_clause_cols)}
                    from {dataset.dataset_table_name}
                    where {numeric_col} is not null
                    group by {", ".join(all_group_by_cols)}
                """

        results = ddb_conn.sql(query).df()

        series = self.group_query_results_to_numeric_metrics(
            results,
            "sum",
            segmentation_cols,
            "ts",
        )
        # preserve dimension that identifies the name of the numeric column used for the aggregation
        for point in series:
            point.dimensions.append(
                Dimension(name="column_name", value=unescape_identifier(numeric_col)),
            )

        metric = self.series_to_metric(self.METRIC_NAME, series)
        return [metric]
